
/* SSD1306Wire.h functions, see https://github.com/ThingPulse/esp8266-oled-ssd1306

  disply.clear()        - clear the local pixel buffer
  disply.display()      - write the buffer to the display
  disply.displayOff()   - turn the display OFF
  disply.displayOn()    - turn the display ON
  display.drawCircle(int16_t x, int16_t y, int16_t radius)
  display.drawFastImage(int16_t x, int16_t y, int16_t width, int16_t height, const uint8_t *image)
  display.drawHorizontalLine(int16_t x, int16_t y, int16_t length)
  display.drawLine(int16_t x0, int16_t y0, int16_t x1, int16_t y1)  - draw from x0k,y0 to x1,y1
  display.drawProgressBar(uint16_t x, uint16_t y, uint16_t width, uint16_t height, uint8_t progress)
  display.drawRect(int16_t x, int16_t y, int16_t width, int16_t height)
  display.drawString(int16_t x, int16_t y, String text)
  display.drawStringMaxWidth(int16_t x, int16_t y, int16_t maxLineWidth, String text)
  display.drawVerticalLine(int16_t x, int16_t y, int16_t length)
  display.drawXbm(int16_t x, int16_t y, int16_t width, int16_t height, const char* xbm)
  display.fillCircle(int16_t x, int16_t y, int16_t radius)
  display.fillRect(int16_t x, int16_t y, int16_t width, int16_t height)
  display.flipScreenVertically()   - flip the current screen upside down
  disply.init()         - initialise the display
  disply.invertDisplay()- inverted display mode
  disply.end()          - free memory used by the display
  disply.displayOn()    - turn the display ON
  display.mirrorScreen() - draw the screen mirrored
  disply.normalDisplay()- normal display mode
  disply.reconnect()    - connect again through I2C
  display.setBrightness(uint8_t)
  display.setColor(OLEDDISPLAY_COLOR color) - BLACK, WHITE, INVERSE
  display.setContrast(uint8_t contrast, uint8_t precharge = 241, uint8_t comdetect = 64)
  display.setFont(const uint8_t* fontData)
  display.setPixel(int16_t x, int16_t y)  - set a pixel at x,y
  display.setTextAlignment(OLEDDISPLAY_TEXT_ALIGNMENT textAlignment)  - TEXT_ALIGN_LEFT, TEXT_ALIGN_CENTER, TEXT_ALIGN_RIGHT, TEXT_ALIGN_CENTER_BOTH

  W = display.getStringWidth(const char* text, uint16_t length)
  W = display.getStringWidth(String text)
  
  Display size is 128 x 64 (w x h) pixels
  Default fonts are ArialMT at 10, 16 & 24 pixels high
  Text options include:
    2 x 24, + 1 x 10 pnt lines
    1 x 24, + 2 x 16 pnt lines
    1 x 24, + 3 x 10 pnt lines
    4 x 16 pnt lines, no line spaces, decending characters will merge
    3 x 16, + 1 x 10 pnt lines
    6 x 10 pnt lines, no line spaces, decending characters will merge
*/

// --------------------------------------------------------------------------------

void Display_40ms() {
  // called every 40ms from the main loop
  if (DispNOP) {return;}  // this function is aborted, no I2C activity
  
//  if (DispDel > 0) {DispDel--; return;} // don't display if DispDel > 0
  if (DispClr) {display.clear(); DispClr = false; return;}
  if (DispDisp) {display.display(); DispDisp = false; DispClr = true; return;}

  // we refresh individual displays based on a counter, DispCnt
  // set DispCnt -ve to prevent the display update process
  DispCnt--;
  if (DispCnt < 1) {
    // DispCnt is reset for display updates, so can be different for each
    if (DispBack) {
      // draw a back arrow on the screen
      display.clear();
      display.drawLine(48,32,80,16);
      display.drawLine(48,32,80,48);
      display.drawLine(80,16,80,48);
      DispCnt = 10;
    } else {
      if (DispNow > 0) {
        // special display modes used in calibration etc.
        switch(DispNow) {
          case 1: DispCnt = 3; Display_ServoVals(); break;
        }
      } else {
        // normal menu screens, display depends on menu modes
        switch(DispMenu) {
          case 0: Display_40ms_Sensors(); break;
          case 1: Display_40ms_Tasks(); break;
          case 2: Display_40ms_Sumo(); break;
        }
      }
    }
    DispDisp = true; // having created the display contents, set flag to display it
    // the following line puts a rectangle around every display, showing the boundry region
//    display.drawRect(0,0,128,64);
  }
}

// --------------------------------------------------------------------------------

void Display_40ms_Sensors() {
  // DispMenu == 0 so display sensor screens
  switch (DispPnt) {
    case 0: // Battery screens
      switch(DispMode) {
        case 0: DispCnt = 10; Display_Battery(); break;
        case 1: DispCnt = 10; Display_BatData(); break;
        default: DispMode = 0; break; // over-run, this should not be necessary
      } break;
    case 1: // MPU6050 screens
      switch(DispMode) {
        case 0: DispCnt = 3; Display_MPU_Levels(); break;
        case 1: DispCnt = 3; Display_MPU_Rot(); break;
        case 2: DispCnt = 5; Display_MPU_Acc(); break;
        case 3: DispCnt = 5; Display_MPU_Gyr(); break;
        case 4: DispCnt = 5; Display_MPU_Ang(); break;
        default: DispMode = 0; break; // over-run, this should not be necessary
      } break;
    case 2: // LTOF ranger screens
      switch(DispMode) {
        case 0: // display LTOF inactive
          if (DispChng) {
            if (LTOF_On > 0) {VL53L0X_OFF();}
            ServoTaskPnt = -1;  // centre servo
          } DispCnt = 10; Display_LTOF(); break;
        case 1: // display LTOF range values
          if (DispChng) {
            if (LTOF_On < 1) {VL53L0X_ON();}
            ServoTaskPnt = -1;  // centre servo
          } DispCnt = 3; Display_LTOF(); break;
        case 2: 
          if (DispChng) {
            if (LTOF_On < 1) {VL53L0X_ON();}
            LED_Task = -30; ServoTaskPnt = -2;  // initiate scanning
            ServoAuto = false;
          } DispCnt = 3; Display_LTOF_Data(); break;
        case 3: 
          if (DispChng) {
            if (LTOF_On < 1) {VL53L0X_ON();}
            LED_Task = -30; ServoTaskPnt = -2;  // initiate scanning
            ServoAuto = true;
          } DispCnt = 3; Display_LTOF_TgtData(); break;
        case 4: 
          if (DispChng) {
            if (LTOF_On < 1) {VL53L0X_ON();}
            LED_Task = -30; ServoTaskPnt = -2;  // initiate scanning
            ServoAuto = true;
          } DispCnt = 3; Display_LTOF_TgtPoint(); break;
        default: DispMode = 0; break; // over-run, this should not be necessary
      } break;
    case 3: // WiFi screens
      switch(DispMode) {
        case 0: DispCnt = 5; Display_WiFiStatus(); break;
        case 1: DispCnt = 5; Display_Wii_Ctrl(); break;
        default: DispMode = 0; break; // over-run, this should not be necessary
      } break;
    case 4: // IR sensor screens
      if (DispChng) {
        if (DispMode == 1) {initialiseIR();}
        if (DispMode > 2) {clearADCdata();}
      }
      switch(DispMode) {
        case 0: DispCnt = 3; Display_IRSensors(); break;
        case 1: DispCnt = 9; Display_IRLimits(); break;
        case 2: DispCnt = 3; Display_IRSensors100P(); break;
        case 3: DispCnt = 3; ADC_Rec = 0; Display_IRdata(); break;
        case 4: DispCnt = 3; ADC_Rec = 1; Display_IRdata(); break;
        case 5: DispCnt = 3; ADC_Rec = 2; Display_IRdata(); break;
        case 6: DispCnt = 3; ADC_Rec = 3; Display_IRdata(); break;
        case 7: DispCnt = 3; ADC_Rec = 4; Display_IRdata(); break;
        default: DispMode = 0; break; // over-run, this should not be necessary
      } break;
  }
  DispChng = false;
}

// --------------------------------------------------------------------------------

void Display_40ms_Sumo() {
  // DispMenu == 2 so display Sumo screens
  switch (DispPnt) {
    case 0:
      switch(DispMode) {
        case 0: DispCnt = 50; Display_Sumo_Mode0(); break;
      } break;
  }
  DispChng = false;
}

// --------------------------------------------------------------------------------

void Display_40ms_Tasks() {
  // DispMenu == 1 so display task screens
  switch (DispPnt) {
    case 0:
      switch(DispMode) {
        case 0: DispCnt = 50; Display_Task_Drive(); break;
      } break;
    case 1:
      switch(DispMode) {
        case 0: DispCnt = 50; Display_Task_Hunt(); break;
      } break;
    case 2:
      switch(DispMode) {
        case 0: DispCnt = 50; Display_Task_Tip(); break;
      } break;
    case 3:
      switch(DispMode) {
        case 0: DispCnt = 50; Display_Task_Target(); break;
      } break;
    case 4:
      switch(DispMode) {
        case 0: DispCnt = 50; Display_Task_Distance(); break;
      } break;
  }
  DispChng = false;
}

// --------------------------------------------------------------------------------

void Display_Battery() {
  // display battery information
  display.setFont(ArialMT_Plain_16);
  display.setTextAlignment(TEXT_ALIGN_CENTER);
  display.drawString(64, 06, "Battery");
  if (BatAvg > BatCritical) {
    // battery powered mode
    display.setTextAlignment(TEXT_ALIGN_LEFT);
    display.drawString(0, 27, String(BatAvg));
    display.setTextAlignment(TEXT_ALIGN_CENTER);
    display.drawString(64, 27, String(float(BatAvg)/BatCal) + "v");
    display.setTextAlignment(TEXT_ALIGN_RIGHT);
    AnyLong = (BatAvg - BatCritical)*100/(BatMax - BatCritical);
    if (AnyLong > 100) {AnyLong = 100;} // limit max to 100%
    if (AnyLong < 0) {AnyLong = 0;}     // limit min to 0%
    display.drawString(127, 27, String(AnyLong) + "%");
    // display the time
    display.setTextAlignment(TEXT_ALIGN_LEFT);
    display.drawString(0, 48, getHmsTime(millis()));
    display.setTextAlignment(TEXT_ALIGN_RIGHT);
    display.drawString(127, 48, BatTime$);
  } else {
    // USB powered mode
    display.setTextAlignment(TEXT_ALIGN_CENTER);
    display.drawString(64, 27, "USB Pwr");
    // display the time
    display.drawString(64, 48, getHmsTime(millis()));
  }
}

// --------------------------------------------------------------------------------

void Display_BatData() {
  // plot the BatData[] array as a rolling average
  DispCX = 4;
  for (int zB = 0; zB < 60; zB++) {
    if (BatData[zB] > 0) {display.drawRect(DispCX,63 - BatData[zB],2,BatData[zB]);}
    DispCX += 2;
  }
  // draw axice
  DispCX = 4;
  display.drawLine(DispCX,0,DispCX,63);
  display.drawLine(DispCX,63,DispCX + 118,63);
}

// --------------------------------------------------------------------------------

void Display_Count(int zNum) {
  // displays a number in the centre of the screen, normally a countdown
  // locks display updates to prevent display task from overwriting count
  display.clear();
  display.setFont(ArialMT_Plain_24);
  display.setTextAlignment(TEXT_ALIGN_CENTER);
  display.drawString(64, 20, String(zNum));
  display.display();
  DispClr = true; // display intro, then clear memory ready for first screen
  DispNOP = true; // prevent display updates
}

// --------------------------------------------------------------------------------

void Display_Init() {
  // initialise the OLED display
  display.init();
  display.clear();
  display.setContrast(255);
//  display.flipScreenVertically();
  display.display();
  DispClr = false; DispDisp = false;
}

// --------------------------------------------------------------------------------

void Display_Intro() {
  // first screen to be displayed
  // default fonts are ArialMT_Plain_10, ArialMT_Plain_16 and ArialMT_Plain_24
  display.setFont(ArialMT_Plain_24);
  display.setTextAlignment(TEXT_ALIGN_CENTER);
  display.drawString(64, 4, "SUMO");
  display.drawString(64, 36, "ROBOT");
  display.display();
  DispClr = true;  // display intro, then clear memory ready for first screen
}

// --------------------------------------------------------------------------------

void Display_IRLimits() {
  // display the IR sensor threshold limit values  
  display.setFont(ArialMT_Plain_16);
  display.setTextAlignment(TEXT_ALIGN_CENTER);
  display.drawString(64, 22, "SENSORS");
  display.setFont(ArialMT_Plain_10);
  display.drawString(64,  0, String(AvBlk[1]));
  display.drawString(64, 10, String(AvWht[1]));
  display.setTextAlignment(TEXT_ALIGN_LEFT);
  display.drawString(0,  4, String(AvBlk[0]));
  display.drawString(0, 14, String(AvWht[0]));
  display.drawString(0, 44, String(AvBlk[4]));
  display.drawString(0, 54, String(AvWht[4]));
  display.setTextAlignment(TEXT_ALIGN_RIGHT);
  display.drawString(127,  4, String(AvBlk[2]));
  display.drawString(127, 14, String(AvWht[2]));
  display.drawString(127, 44, String(AvBlk[3]));
  display.drawString(127, 54, String(AvWht[3]));
  display.setTextAlignment(TEXT_ALIGN_CENTER);
  display.drawString(64, 43, getMaskStr());
  if (IR_BaseF) {display.drawString(64, 54, "WHITE");}
  else {display.drawString(64, 54, "BLACK");}
}

// --------------------------------------------------------------------------------

void Display_IRSensors() {
  // display IR sensor raw values 
  display.setFont(ArialMT_Plain_16);
  display.setTextAlignment(TEXT_ALIGN_CENTER);
  display.drawString(64, 21, "SENSORS");
  display.drawString(64, 0, String(AvRaw[1]));    // top centre / front centre
  display.setTextAlignment(TEXT_ALIGN_LEFT);
  display.drawString(0, 4, String(AvRaw[0]));     // top left / front left
  display.drawString(0, 48, String(AvRaw[4]));    // bottom left / rear left
  display.setTextAlignment(TEXT_ALIGN_RIGHT);
  display.drawString(127, 4, String(AvRaw[2]));   // top right / front right
  display.drawString(127, 48, String(AvRaw[3]));  // bottom right / rear right
  display.setFont(ArialMT_Plain_10);
  display.setTextAlignment(TEXT_ALIGN_CENTER);
  display.drawString(64, 43, getMaskStr());
  if (IR_BaseF) {display.drawString(64, 54, "WHITE");}
  else {display.drawString(64, 54, "BLACK");}
}

// --------------------------------------------------------------------------------

void Display_IRSensors100P() {
  // display IR sensor values as percentages 
  display.setFont(ArialMT_Plain_16);
  display.setTextAlignment(TEXT_ALIGN_CENTER);
  display.drawString(64, 21, "SENSORS");
  display.drawString(64, 0, String(Av100[1]) + "%");    // top centre / front centre
  display.setTextAlignment(TEXT_ALIGN_LEFT);
  display.drawString(0, 4, String(Av100[0]) + "%");     // top left / front left
  display.drawString(0, 48, String(Av100[4]) + "%");    // bottom left / rear left
  display.setTextAlignment(TEXT_ALIGN_RIGHT);
  display.drawString(127, 4, String(Av100[2]) + "%");   // top right / front right
  display.drawString(127, 48, String(Av100[3]) + "%");  // bottom right / rear right
  display.setFont(ArialMT_Plain_10);
  display.setTextAlignment(TEXT_ALIGN_CENTER);
  display.drawString(64, 43, getMaskStr());
  if (IR_BaseF) {display.drawString(64, 54, "WHITE");}
  else {display.drawString(64, 54, "BLACK");}
}

// --------------------------------------------------------------------------------

void Display_IRdata() {
  // display a sensors readings on a timeline
  // there are 40 data points so we can use 3 pixels per point on the X axis
  int zCx = 4;              // x-axis start point
  int zVP = ADC_RecPnt + 1; // oldest data point stored
  display.setFont(ArialMT_Plain_16);
  display.setTextAlignment(TEXT_ALIGN_LEFT);
  for (int zP = 0; zP < 40; zP++) {
    if (zVP > 39) {zVP = 0;}
    int zCy = ADC_Data[zVP]/50;
    if (zCy > 0) {display.fillRect(zCx,63-zCy,3,zCy);}
    else if (zCy == 0) {display.fillRect(zCx, 63,3,1);}
    zVP++; zCx += 3;
  }
  display.setColor(INVERSE);
  switch(ADC_Rec) {
    case 0: display.drawString(0, 0, "F-L"); break; // front left
    case 1: display.drawString(0, 0, "F-C"); break; // front centre
    case 2: display.drawString(0, 0, "F-R"); break; // front right
    case 3: display.drawString(0, 0, "R-R"); break; // rear right
    case 4: display.drawString(0, 0, "R-L"); break; // rear left
  }
  display.setColor(WHITE);
}

// --------------------------------------------------------------------------------

void Display_LTOF() {
  // display the range value from the LTOF if it is enabled
  display.setFont(ArialMT_Plain_16);
  display.setTextAlignment(TEXT_ALIGN_CENTER);

  if (LTOF_On < 1) {
    // LTOF is not active
    display.drawString(64, 6, "LTOF");
    display.drawString(64,42, "Disabled");
  } else {
    // LTOF is active
    display.drawString(64, 0, "LTOF");
    display.setTextAlignment(TEXT_ALIGN_RIGHT);
    
    display.drawString(62, 16, "Range:");
    display.setTextAlignment(TEXT_ALIGN_LEFT);
    display.drawString(66, 16, String(Range));
    
    display.setTextAlignment(TEXT_ALIGN_RIGHT);
    display.drawString(62, 32, "Filter:");
    display.setTextAlignment(TEXT_ALIGN_LEFT);
    display.drawString(66, 32, String(RangeFtd));
    
    display.setTextAlignment(TEXT_ALIGN_RIGHT);
    display.drawString(62, 48, "FPS:");
    display.setTextAlignment(TEXT_ALIGN_LEFT);
    display.drawString(66, 48, String(LTOF_fps));
  }
}

// --------------------------------------------------------------------------------

void Display_LTOF_Data() {
  // displays the data stored in the RangeData array
  // the table consists of 61 range values, from RangeMin to RangeOL
  int zMax = 0;
  int zMin = RangeMax;
  int zDMax = 1 + (ServoAngMaxDef - ServoAngMinDef);
  DispCX = 62 - (zDMax/2); int zXMin = DispCX;
  for (int zD = 0; zD < zDMax; zD++) {
    int zY = map(RangeData[zD],RangeMin,RangeMax,63,0);
    if (RangeData[zD] < zMin) {zMin = RangeData[zD]; zXMin = DispCX;}
    else if (RangeData[zD] > zMax) {zMax = RangeData[zD];}
    display.drawLine(DispCX,63,DispCX,zY);
    DispCX++;
  }
  // draw the min range pointer
  display.setColor(INVERSE);
  display.drawLine(zXMin-1,62,zXMin-1,63);
  display.drawLine(zXMin,61,zXMin,63);
  display.drawLine(zXMin+1,62,zXMin+1,63);

  display.setColor(WHITE);
  display.setFont(ArialMT_Plain_10);
  display.setTextAlignment(TEXT_ALIGN_LEFT);
  display.drawString(0, 0, "Max");
  display.drawString(0, 12, String(zMax));
  display.drawString(0, 41, "Min");
  display.drawString(0, 53, String(zMin));
  
  display.setTextAlignment(TEXT_ALIGN_RIGHT);
  display.drawString(127, 0, "Tgt ");
  display.drawString(127, 12, String(ServoAngTgtAvg) + "°");
  display.drawString(127, 41, "Min ");
  display.drawString(127, 53, String(ServoAngTgt) + "°");
}

// --------------------------------------------------------------------------------

void Display_LTOF_TgtData() {
  // displays the data stored in the RangeData array, filtered by
  // the lower and upper 10mm cut-off points
  // the table consists of 61 range values, from RangeMin to RangeOL
  int zMax = 0;           // max data value
  int zMin = RangeMax;    // min data value
  int zDA = ServoAngMin;  // data point angle
  int zDMax = 1 + (ServoAngMaxDef - ServoAngMinDef);
  int zRD = RangeMax;     // modified range data varaible
  DispCX = 62 - (zDMax/2); int zXMin = DispCX;
  for (int zD = 0; zD < zDMax; zD++) {
    zRD = RangeData[zD]; zDA = zD + ServoAngMinDef;
//    if ((zDA < ServoAngTgtLL) || (zDA > ServoAngTgtUL)) {zRD = RangeMax;}
    int zY = map(zRD,RangeMin,RangeMax,63,0);
    if (RangeData[zD] < zMin) {zMin = RangeData[zD]; zXMin = DispCX;}
    else if (RangeData[zD] > zMax) {zMax = RangeData[zD];}
    display.drawLine(DispCX,63,DispCX,zY);
    DispCX++;
  }
  // draw the min range pointer
  display.setColor(INVERSE);
  display.drawLine(zXMin-1,62,zXMin-1,63);
  display.drawLine(zXMin,61,zXMin,63);
  display.drawLine(zXMin+1,62,zXMin+1,63);

  display.setColor(WHITE);
  display.setFont(ArialMT_Plain_10);
  display.setTextAlignment(TEXT_ALIGN_LEFT);
  display.drawString(0, 0, "Max");
  display.drawString(0, 12, String(zMax));
  display.drawString(0, 41, "Min");
  display.drawString(0, 53, String(zMin));
  
  display.setTextAlignment(TEXT_ALIGN_RIGHT);
  display.drawString(127, 0, "Tgt ");
  display.drawString(127, 12, String(ServoAngTgtAvg) + "°");
  display.drawString(127, 41, "Min ");
  display.drawString(127, 53, String(ServoAngTgt) + "°");
}

// --------------------------------------------------------------------------------

void Display_LTOF_TgtPoint() {
  // displays the target position and range with a direction pointer
  // 90 Deg = 1.5708 radians
  // draw the angle pointer
  if (ServoAngTgtAvg > 90) {
    DispCX = 63 - int(63.0 * cos(0.0174533 * float(ServoAngTgtAvg)));
    DispCY = 63 - int(63.0 * cos(0.0174533 * float(ServoAngTgtAvg - 90)));
  } else if (ServoAngTgtAvg < 90) {
    DispCX = 63 - int(63.0 * cos(0.0174533 * float(ServoAngTgtAvg)));
    DispCY = 63 - int(63.0 * cos(0.0174533 * float(90 - ServoAngTgtAvg)));
  } else {
    DispCX = 64; DispCY = 0;
  }
//  PrintTx += "\nServoAngTgtAvg " + String(ServoAngTgtAvg) + "\n";
//  PrintTx += "DispCX " + String(DispCX) + "\tDispCY " + String(DispCY) + "\n";
  display.drawLine(64,63,DispCX,DispCY);

  // draw the target at range
  display.setColor(INVERSE);
  float zRange = (63.0 * RangeFtdTgt)/RangeMax;
  if (ServoAngTgtAvg > 90) {
    DispCX = 63 - int(zRange * cos(0.0174533 * float(ServoAngTgtAvg)));
    DispCY = 63 - int(zRange * cos(0.0174533 * float(ServoAngTgtAvg - 90)));
  } else if (ServoAngTgtAvg < 90) {
    DispCX = 63 - int(zRange * cos(0.0174533 * float(ServoAngTgtAvg)));
    DispCY = 63 - int(zRange * cos(0.0174533 * float(90 - ServoAngTgtAvg)));
  } else {
    DispCX = 64; DispCY = 63 - zRange;
  }
  display.drawCircle(DispCX,DispCY,4);

  SpeedTgt = 1000; Task_Target_Steer(); // see the effect on normalised steering
//  display.setColor(WHITE);
  display.setFont(ArialMT_Plain_10);
  display.setTextAlignment(TEXT_ALIGN_LEFT);
  display.drawString(2,  0, "L-Steer");
  display.drawString(2, 36, "Angle");
  display.setFont(ArialMT_Plain_16);
  display.drawString(0, 11, String(L_SpdTgt));
  display.drawString(0, 47, String(ServoAngTgtAvg) + "°");
  display.setTextAlignment(TEXT_ALIGN_RIGHT);
  display.setFont(ArialMT_Plain_10);
  display.drawString(127,  0, "R-Steer");
  display.drawString(127, 36, "Range");
  display.setFont(ArialMT_Plain_16);
  display.drawString(125, 11, String(R_SpdTgt));
  display.drawString(125, 47, String(RangeFtdTgt));
  display.setColor(WHITE);
}

// --------------------------------------------------------------------------------

void Display_MPU_Acc() {
  // display MPU6050 accelerometer information
  display.setFont(ArialMT_Plain_16);
  display.setTextAlignment(TEXT_ALIGN_CENTER);
  display.drawString(64, 0, "MPU Acc.");

  display.setTextAlignment(TEXT_ALIGN_LEFT);
  display.drawString(0, 16, "AcX");
  display.setTextAlignment(TEXT_ALIGN_LEFT);
  display.drawString(36, 16, String(AcX));
  display.setTextAlignment(TEXT_ALIGN_RIGHT);
  display.drawString(127, 16, String(AcXAvg));

  display.setTextAlignment(TEXT_ALIGN_LEFT);
  display.drawString(0, 32, "AcY");
  display.setTextAlignment(TEXT_ALIGN_LEFT);
  display.drawString(36, 32, String(AcY));
  display.setTextAlignment(TEXT_ALIGN_RIGHT);
  display.drawString(127, 32, String(AcYAvg));

  display.setTextAlignment(TEXT_ALIGN_LEFT);
  display.drawString(0, 48, "AcZ");
  display.setTextAlignment(TEXT_ALIGN_LEFT);
  display.drawString(36, 48, String(AcZ));
  display.setTextAlignment(TEXT_ALIGN_RIGHT);
  display.drawString(127, 48, String(AcZAvg));
}

// --------------------------------------------------------------------------------

void Display_MPU_Ang() {
  // display MPU6050 angle information
  display.setFont(ArialMT_Plain_16);
  display.setTextAlignment(TEXT_ALIGN_CENTER);
  display.drawString(64, 0, "MPU Angles");

  display.setTextAlignment(TEXT_ALIGN_LEFT);
  display.drawString(0, 16, "Pitch");
  display.drawString(40, 16, String(PitchGyr,1) + "°");
  display.setTextAlignment(TEXT_ALIGN_RIGHT);
  display.drawString(127, 16, String(PitchAcc,1) + "°");

  display.setTextAlignment(TEXT_ALIGN_LEFT);
  display.drawString(0, 32, "Yaw");
  display.drawString(40, 32, String(YawGyr,1) + "°");
  display.setTextAlignment(TEXT_ALIGN_RIGHT);
  display.drawString(127, 32, String(YawAcc,1) + "°");

  display.setTextAlignment(TEXT_ALIGN_LEFT);
  display.drawString(0, 48, "Rot.");
  display.drawString(40, 48, String(RotGyr,1) + "°");
}

// --------------------------------------------------------------------------------

void Display_MPU_Gyr() {
  // display MPU6050 gyroscope information
  display.setFont(ArialMT_Plain_16);
  display.setTextAlignment(TEXT_ALIGN_CENTER);
  display.drawString(64, 0, "MPU Gyros");

  display.setTextAlignment(TEXT_ALIGN_LEFT);
  display.drawString(0, 16, "GyX");
  display.drawString(40, 16, String(GyX));

  display.drawString(0, 32, "GyY");
  display.drawString(40, 32, String(GyY));

  display.drawString(0, 48, "GyZ");
  display.drawString(40, 48, String(GyZ));

  if (GyCal < 1) {
    display.setTextAlignment(TEXT_ALIGN_RIGHT);
    display.drawString(127, 16, String(GyXAvg));
    display.drawString(127, 32, String(GyYAvg));
    display.drawString(127, 48, String(GyZAvg));
  }
}

// --------------------------------------------------------------------------------

void Display_MPU_Levels() {
  // display MPU6050 angles like s spirit level
  display.setFont(ArialMT_Plain_10);

  if ((fabs(PitchGyr) <= 2.0) && (fabs(YawGyr) <= 2.0)) {
    // angles are <= 2.0°
    display.setTextAlignment(TEXT_ALIGN_LEFT);
    display.drawString(72, 0, String(PitchGyr,1) + "°");
    display.setTextAlignment(TEXT_ALIGN_RIGHT);
    display.drawString(24,27, String(YawGyr,1) + "°");
    display.drawLine(32,32,96,32);    // horizontal centre line
    DispCX = 62 + (16 * YawGyr);      // Yaw bubble
    display.drawRect(DispCX,30,5,5);  // horizontal bubble
    display.drawLine(64,0,64,63);     // vertical centre line
    DispCY = 30 - (16 * PitchGyr);    // Pitch bubble
    display.drawRect(62,DispCY,5,5);  // vertical bubble
  } else {
    // angles are <= 10.0°
    if (YawGyr >= 0.0) {
      DispCX = 96; DispCY = 2 * fabs(YawGyr);
      display.drawLine(DispCX,32 - DispCY,DispCX,32 + DispCY);  // horizontal wedge right
      display.drawLine(32,32,DispCX,32 - DispCY);
      display.drawLine(32,32,DispCX,32 + DispCY);
      display.setTextAlignment(TEXT_ALIGN_LEFT);
      display.drawString(DispCX + 4,27, String(YawGyr,1) + "°");
    } else {
      DispCX = 32; DispCY = 2 * fabs(YawGyr);
      display.drawLine(DispCX,32 - DispCY,DispCX,32 + DispCY);  // horizontal wedge left
      display.drawLine(96,32,DispCX,32 - DispCY);
      display.drawLine(96,32,DispCX,32 + DispCY);
      display.setTextAlignment(TEXT_ALIGN_RIGHT);
      display.drawString(DispCX - 4,27, String(YawGyr,1) + "°");
    }
    if (PitchGyr >= 0.0) {
      DispCX = 2 * fabs(PitchGyr); DispCY = 0; 
      display.drawLine(64 - DispCX,DispCY,64 + DispCX,DispCY);  // vertical wedge up
      display.drawLine(64,64,64 - DispCX,DispCY);
      display.drawLine(64,64,64 + DispCX,DispCY);
      if (YawGyr < 0.0) {
        display.setTextAlignment(TEXT_ALIGN_LEFT);
        display.drawString(70,53, String(PitchGyr,1) + "°");
      } else {
        display.setTextAlignment(TEXT_ALIGN_RIGHT);
        display.drawString(58,53, String(PitchGyr,1) + "°");
      }
    } else {
      DispCX = 2 * fabs(PitchGyr); DispCY = 63; 
      display.drawLine(64 - DispCX,DispCY,64 + DispCX,DispCY);  // vertical wedge down
      display.drawLine(64,0,64 - DispCX,DispCY);
      display.drawLine(64,0,64 + DispCX,DispCY);
      if (YawGyr < 0.0) {
        display.setTextAlignment(TEXT_ALIGN_LEFT);
        display.drawString(70,0, String(PitchGyr,1) + "°");
      } else {
        display.setTextAlignment(TEXT_ALIGN_RIGHT);
        display.drawString(58,0, String(PitchGyr,1) + "°");
      }
    }
  }
}

// --------------------------------------------------------------------------------

void Display_MPU_Rot() {
  // display MPU6050 rotation angle
  display.setFont(ArialMT_Plain_16);  // fonts are ArialMT_Plain_10, ArilMT_Plain_24
  display.setTextAlignment(TEXT_ALIGN_CENTER);
  display.drawString(64, 24, String(RotGyr,1) + "°");
  display.setFont(ArialMT_Plain_10);  // fonts are ArialMT_Plain_10, ArilMT_Plain_24
  display.setTextAlignment(TEXT_ALIGN_LEFT);
  display.drawString(0, 54, String(RotDif,3) + "°");
  display.setTextAlignment(TEXT_ALIGN_RIGHT);
  display.drawString(127, 54, String(GyYOff));
  
  display.drawCircle(64,32,31);
  display.drawLine(21,32,30,32);
  display.drawLine(97,32,106,32);
  if (RotGyr > 0.0) {
    display.drawLine(21,32,26,21); display.drawLine(26,21,30,32);
    display.drawLine(97,32,102,41); display.drawLine(102,41,106,32);
  } else if (RotGyr < 0.0) {
    display.drawLine(21,32,26,41); display.drawLine(26,41,30,32);
    display.drawLine(97,32,102,21); display.drawLine(102,21,106,32);
  }
}

// --------------------------------------------------------------------------------

void Display_Overlay_Sensors() {
  // called when changing menu modes
  display.clear();
  display.drawRect(16,20,96,25);
  display.setFont(ArialMT_Plain_16);  // fonts are ArialMT_Plain_10, ArilMT_Plain_24
  display.setTextAlignment(TEXT_ALIGN_CENTER);
  display.drawString(64, 24, "SENSORS");
  display.setColor(WHITE);
  display.display();
}

// --------------------------------------------------------------------------------

void Display_Overlay_Sumo() {
  // called when changing menu modes
  display.clear();
  display.drawRect(24,20,80,25);
  display.setFont(ArialMT_Plain_16);  // fonts are ArialMT_Plain_10, ArilMT_Plain_24
  display.setTextAlignment(TEXT_ALIGN_CENTER);
  display.drawString(64, 24, "SUMO");
  display.setColor(WHITE);
  display.display();
}

// --------------------------------------------------------------------------------

void Display_Overlay_Tasks() {
  // called when changing menu modes
  display.clear();
  display.drawRect(24,20,80,25);
  display.setFont(ArialMT_Plain_16);  // fonts are ArialMT_Plain_10, ArilMT_Plain_24
  display.setTextAlignment(TEXT_ALIGN_CENTER);
  display.drawString(64, 24, "TASKS");
  display.setColor(WHITE);
  display.display();
}

// --------------------------------------------------------------------------------

void Display_QText2x16(String zT0$,String zT1$) {
  // called to display two lines of text + <NO OK>, centre screen in font 16
  display.clear();
  display.setFont(ArialMT_Plain_16);  // fonts are ArialMT_Plain_10,16,24
  display.setTextAlignment(TEXT_ALIGN_CENTER);
  display.drawString(64, 4, zT0$);
  display.drawString(64, 24, zT1$);
  display.setTextAlignment(TEXT_ALIGN_LEFT);
  display.drawString(0, 44, "< NO");
  display.setTextAlignment(TEXT_ALIGN_RIGHT);
  display.drawString(127, 44, "OK >");
  display.display();
  display.clear(); DispNOP = true;
}

// --------------------------------------------------------------------------------

void Display_Reboot() {
  // called when a soft reset is about to be initiated
  display.clear();
  display.setFont(ArialMT_Plain_24);  // fonts are ArialMT_Plain_10, ArilMT_Plain_24
  display.setTextAlignment(TEXT_ALIGN_CENTER);
  display.drawString(64, 10, "RESET");
  display.setFont(ArialMT_Plain_16);
  display.drawString(64, 40, "Pending...");
  display.display(); DispNOP = true;
}

// --------------------------------------------------------------------------------

void Display_ServoVals() {
  // used in cdalibratio mode to display servo and filtered LTOF values 
  display.setFont(ArialMT_Plain_16);
  display.setTextAlignment(TEXT_ALIGN_RIGHT);
  display.drawString(63,  6, "PWM:");
  display.drawString(63, 24, "Ang:");
  display.drawString(63, 42, "LTOF:");
  display.setTextAlignment(TEXT_ALIGN_LEFT);
  display.drawString(65,  6, String(servoVal));
  display.drawString(65, 24, String(ServoAng));
  display.drawString(65, 42, String(RangeFtd));
}

// --------------------------------------------------------------------------------

void Display_Sumo_Mode0() {
  // display Sumo Mode0 screen  
  display.setFont(ArialMT_Plain_16);
  display.setTextAlignment(TEXT_ALIGN_CENTER);
  display.drawString(64, 12, "SUMO M0");
  display.setFont(ArialMT_Plain_10);
  display.drawString(64, 36, "< PRESS SW0 >");
}

// --------------------------------------------------------------------------------

void Display_Task_Distance() {
  // display DISTANCE Task  
  display.setFont(ArialMT_Plain_16);
  display.setTextAlignment(TEXT_ALIGN_CENTER);
  display.drawString(64, 12, "DISTANCE TEST");
  display.setFont(ArialMT_Plain_10);
  display.drawString(64, 36, "< PRESS SW0 >");
}

// --------------------------------------------------------------------------------

void Display_Task_Drive() {
  // display Hunt Task  
  display.setFont(ArialMT_Plain_16);
  display.setTextAlignment(TEXT_ALIGN_CENTER);
  display.drawString(64, 12, "DRIVE TASK");
  display.setFont(ArialMT_Plain_10);
  display.drawString(64, 36, "< PRESS SW0 >");
}

// --------------------------------------------------------------------------------

void Display_Task_Hunt() {
  // display Hunt Task  
  display.setFont(ArialMT_Plain_16);
  display.setTextAlignment(TEXT_ALIGN_CENTER);
  display.drawString(64, 12, "HUNT TASK");
  display.setFont(ArialMT_Plain_10);
  display.drawString(64, 36, "< PRESS SW0 >");
}

// --------------------------------------------------------------------------------

void Display_Task_Tip() {
  // display Tip Task  
  display.setFont(ArialMT_Plain_16);
  display.setTextAlignment(TEXT_ALIGN_CENTER);
  display.drawString(64, 12, "TIP TEST");
  display.setFont(ArialMT_Plain_10);
  display.drawString(64, 36, "< PRESS SW0 >");
}

// --------------------------------------------------------------------------------

void Display_Task_Target() {
  // display Tip Task  
  display.setFont(ArialMT_Plain_16);
  display.setTextAlignment(TEXT_ALIGN_CENTER);
  display.drawString(64, 12, "TARGET TEST");
  display.setFont(ArialMT_Plain_10);
  display.drawString(64, 36, "< PRESS SW0 >");
}

// --------------------------------------------------------------------------------

void Display_Text2x16(String zT0$,String zT1$) {
  // called to display a two lines of text centre screen in font 16
  display.clear();
  display.setFont(ArialMT_Plain_16);  // fonts are ArialMT_Plain_10,16,24
  display.setTextAlignment(TEXT_ALIGN_CENTER);
  display.drawString(64, 14, zT0$);
  display.drawString(64, 34, zT1$);
  display.display();
  display.clear(); DispNOP = true;
}

// --------------------------------------------------------------------------------

void Display_Title24(String zT$) {
  // called to display a title centre screen
  display.clear();
  display.setFont(ArialMT_Plain_24);  // fonts are ArialMT_Plain_10,16,24
  display.setTextAlignment(TEXT_ALIGN_CENTER);
  display.drawString(64, 20, zT$);
  display.display();
  display.clear(); DispNOP = true;
}

// --------------------------------------------------------------------------------

void Display_WiFiStatus() {
  // displays WiFi link data  
  display.setFont(ArialMT_Plain_16);
  display.setTextAlignment(TEXT_ALIGN_CENTER);
  display.drawString(64, 0, "WiFi STATUS");

  if (WiFiConnected) {
    if (WiFiRxBytes > 0) {
      // some data has been received
      display.setFont(ArialMT_Plain_10);
      display.setTextAlignment(TEXT_ALIGN_RIGHT);
      display.drawString(61, 18, "Rx Bytes:");
      display.setTextAlignment(TEXT_ALIGN_LEFT);
      display.drawString(64, 18, getNumCSV(WiFiRxBytes));
      display.setTextAlignment(TEXT_ALIGN_RIGHT);
      display.drawString(61, 30, "Rx Error:");
      display.setTextAlignment(TEXT_ALIGN_LEFT);
      display.drawString(64, 30, String(WiFiRxErr));
  
      display.setTextAlignment(TEXT_ALIGN_RIGHT);
      display.drawString(61, 41, "Tx Bytes:");
      display.setTextAlignment(TEXT_ALIGN_LEFT);
      display.drawString(64, 41, getNumCSV(WiFiTxBytes));
      display.setTextAlignment(TEXT_ALIGN_RIGHT);
      display.drawString(61, 53, "Tx Error:");
      display.setTextAlignment(TEXT_ALIGN_LEFT);
      display.drawString(64, 53, String(WiFiTxErr));
    } else {
      // transmission errors so display disconnected
      display.setTextAlignment(TEXT_ALIGN_CENTER);
      display.drawString(64, 22, "NO LINK");
      display.setFont(ArialMT_Plain_10);
      display.drawString(64, 48, String(myMAC));
    }
  } else {
    display.drawString(64, 22, "Disconnected");
    display.drawString(64, 44, "Trying " + String(WiFiTryNum));
  }
}

// --------------------------------------------------------------------------------

void Display_Wii_Ctrl() {
  // display 6 bytes received from Wii over ESP-NOW
  display.clear();
  // display I2C values as decimal 0 - 255
  display.setTextAlignment(TEXT_ALIGN_RIGHT);
  display.setFont(ArialMT_Plain_10); DispCX = 58;
  display.drawString(DispCX,  0, String(RxWiFi[0]));
  display.drawString(DispCX, 10, String(RxWiFi[1]));
  display.drawString(DispCX, 20, String(RxWiFi[2]));
  display.drawString(DispCX, 30, String(RxWiFi[3]));
  display.drawString(DispCX, 40, String(RxWiFi[4]));
  display.drawString(DispCX, 50, String(RxWiFi[5]));
  // display I2C values as binary 00000000 - 11111111
  display.setTextAlignment(TEXT_ALIGN_LEFT); DispCX = 66;
  display.drawString(DispCX,  0, GetBIN(RxWiFi[0]));
  display.drawString(DispCX, 10, GetBIN(RxWiFi[1]));
  display.drawString(DispCX, 20, GetBIN(RxWiFi[2]));
  display.drawString(DispCX, 30, GetBIN(RxWiFi[3]));
  display.drawString(DispCX, 40, GetBIN(RxWiFi[4]));
  display.drawString(DispCX, 50, GetBIN(RxWiFi[5]));
  // display 'I2C' in top left corner
  display.setFont(ArialMT_Plain_16);
  display.drawString(2, 0, "Wii");
       if (WiiType == 'C') {display.drawString(2, 18, "Csc");}
  else if (WiiType == 'N') {display.drawString(2, 18, "Nnk");}
}

// --------------------------------------------------------------------------------
