  //##############################################################################
  //
  //  Sumo Robot Move Functions
  //
  //##############################################################################

// these functions are used when controlled by a Wii Nunchuk over WiFi


// ---------------------------------------------------------------------------------

void ClearDistance() {
  // reset the left/right distance variables and set timer
  L_Dst = 0; R_Dst = 0; DistanceMs = millis();
}

// ---------------------------------------------------------------------------------

void getSpeedJXY(int zJXY) {
  // Joystick vector JoyV varies from 8 - 127 which is mapped into a speed profile
  // low to medium demands varies speed between SpdMin and SpdTop, where SpdTop = SpdMid
  // higher demands increase SpdTop up to SpdMax, extending the SpdTop speed range
  // reduced demand lowers SpdTop towards SpdMid
  // Gear changes from 1 - 5 to represent speed from SpdMin - SpdMax
  if (zJXY >= 120) {
    // high demand so increase speed progressively
    if (SpdTop < SpeedMax) {SpdTop += 10;} else {SpdTop = SpeedMax;}
  } else if (zJXY <= 40) {
    // low demand so decrease speed progressively
    if (SpdTop > SpeedMid) {SpdTop -= 10;} else {SpdTop = SpeedMid;}
  }
  
  // limit rate of change of speed to avoid surges in demand
  SpeedTgt = map(zJXY,8,127,SpeedMin,SpdTop);
    if (Speed < SpeedMin) {Speed = SpeedMin;}
  else if (Speed < SpeedTgt) {Speed += 1 + ((SpeedTgt - Speed)/4);}
  else if (Speed > SpeedTgt) {Speed -= 1 + ((Speed - SpeedTgt)/4);}
//  Serial.println("JoyV = " + String(JoyV) + "\tSpeed = " + String(Speed));
}

// ---------------------------------------------------------------------------------

void MoveBkd() {
  // move the robot backwards
  getSpeedJXY(JoyV);
  setWheelSpeed(-Speed,-Speed);
  MoveState = -JoyD;
}

// ---------------------------------------------------------------------------------

void MoveBkdLeft() {
  // move the robot backwards bearing left
  getSpeedJXY(JoyV);
  setWheelSpeed(-Speed/2,-Speed);
  MoveState = -JoyD;
}

// ---------------------------------------------------------------------------------

void MoveBkdRight() {
  // move the robot backwards bearing right
  getSpeedJXY(JoyV);
  setWheelSpeed(-Speed,-Speed/2);
  MoveState = -JoyD;
}

// ---------------------------------------------------------------------------------

void MoveFwd() {
  // move the robot forward
  getSpeedJXY(JoyV);
  setWheelSpeed(Speed,Speed);
  MoveState = JoyD;
}

// ---------------------------------------------------------------------------------

void MoveFwdLft() {
  // move forward turning left
  getSpeedJXY(JoyV);
  setWheelSpeed(Speed/2,Speed);
  MoveState = JoyD;
}

// ---------------------------------------------------------------------------------

void MoveFwdRht() {
  // move forward turning right
  getSpeedJXY(JoyV);
  setWheelSpeed(Speed,Speed/2);
  MoveState = JoyD;
}

// ---------------------------------------------------------------------------------

void MoveSTOP(byte zM) {
  // stop the wheels from turning
  // zM = 0 - servos will stop slowly
  // zM > 0 - servos will be reversed to stop quickly
  if (zM > 0) {
    int zMax = max(int(abs(L_Spd)),int(abs(R_Spd)));
    if (L_Spd > 0) {L_Spd = -200;} else {L_Spd = 200;}
    if (R_Spd > 0) {R_Spd = -200;} else {R_Spd = 200;}
    zMax = map(zMax,150,1000,10,60);  // set reverse period to match current speed
    setWheelSpeed(L_Spd,R_Spd);       // temporarily reverse motors
    delay(zMax);                      // apply speed dependant delay
    setWheelSpeed(0,0);               // now stop the wheels
    synchLoopTimers(zMax);            // add delay to the loop timers
  } else {
    setWheelSpeed(0,0);               // stop the wheels gently
  }
  Speed = 0; SpdTop = SpeedMin; MoveState = 0;
}

// ---------------------------------------------------------------------------------

void MoveTasks() {
  // Entry point for 'Move' tasks, called every 20ms (50Hz) if MoveTask > 0
  if (MoveTask < 0) {
    // -ve tasks initialise variables. Normally used with a new move task
    MoveSub = 0;
  }

  switch(MoveTask) {
    case 1: MoveTask_TiltBack();  break;
    default:
      MoveTask = 0; break;  // catch-all error, stops MoveTask calls
  }
}

// ---------------------------------------------------------------------------------

void MoveTask_TiltBack() {
  // this is the response to the robot being tilted backwards  
}

// ---------------------------------------------------------------------------------

void MoveTurnLeft() {
  // neutral turn to the left
  getSpeedJXY(JoyV);
  if (MoveState == 0) {setWheelSpeed(-Speed,Speed);}
  if (MoveState >= 0) {setWheelSpeed(-Speed/2,Speed/2);}
  else {setWheelSpeed(Speed/2,-Speed/2);}
}

// ---------------------------------------------------------------------------------

void MoveTurnRight() {
  // neutral turn to the right
  getSpeedJXY(JoyV);
  if (MoveState == 0) {setWheelSpeed(Speed,-Speed);}
  else if (MoveState >= 0) {setWheelSpeed(Speed/2,-Speed/2);}
  else {setWheelSpeed(-Speed/2,Speed/2);}
}

// ---------------------------------------------------------------------------------

void setWheelSpeed(int zL_Spd,int zR_Spd) {
  // sends wheel speed values to the two LX-16A servo motors serially @ 115200 baud
  // we also maintain figures for distance travelled for the left and right wheels
  LobotSerialServoSetMode(Serial2, ID1, 1, zL_Spd);
  LobotSerialServoSetMode(Serial2, ID2, 1, -zR_Spd);
  // We use the previous values to determine distance travelled to this point
  // @1000 robot will travel 0.1948mm in 1 millisecond
  // distance is held in a long integer, where 1 == 0.01mm (ie. dist x 100)
  updateDistance();
  L_Spd = zL_Spd; R_Spd = zR_Spd; // record new speed values
}

// ---------------------------------------------------------------------------------

void updateDistance() {
  // We use the current speed values to determine distance travelled to this point
  // @1000 robot will travel 0.1948mm in 1 millisecond
  // distance is held in a long integer, where 1 == 0.0001mm
  // to get distance in mm we have to divide by 10000
  unsigned long zDMs = millis();
  long zDst = map(abs(L_Spd),150,1000,218,1917)*(zDMs - DistanceMs);
  if (L_Spd >= 0) {L_Dst+= zDst;} else {L_Dst-= zDst;}
  zDst = map(abs(R_Spd),150,1000,218,1917)*(zDMs - DistanceMs);
  if (R_Spd >= 0) {R_Dst+= zDst;} else {R_Dst-= zDst;}
//  PrintTx+= "L_Spd " + String(L_Spd) + "\tzDst " + String(zDst) + "\n";
  DistanceMs = zDMs;              // update the distance timer
}

// ---------------------------------------------------------------------------------
