  //##############################################################################
  //
  //  Sumo Robot Main Task Functions
  //
  //##############################################################################


void doMainTasks() {
  // called every 20ms to progress task actions
  // tasks functions can be delayed several cycles by MainDel
  // tasks can be blocked until buttons are released by WaitUp == true
  if (MainDel > 0) {MainDel--; return;} // set > 0 to reduce/delay task rate
  if (WaitUp) {if ((sw0State == LOW) || (sw1State == LOW)) {return;} else {WaitUp = false;}}
  
  switch(MainTask) {
    case  1: Task_AI(); break;
    case  2: Task_Driving(); break;
    case  3: Task_Hunting(); break;
    case  4: Task_TipTest(); break;
    case  5: Task_Target(); break;
    case  6: Task_Distance(); break;
    case 10: Task_SumoM0(); break;
    case 70: Test_SweepRate(); break; // special test, determines servo sweep rate
    case 80: Task_Cal_MPU(); break;
    case 81: Task_Cal_LTOF(); break;
    case 82: Task_Cal_IR(); break;
    case 98:
      // SW0 & SW1 released before invoking MainNext
      if ((sw0State == HIGH) && (sw1State == HIGH)) {MainTask = MainNext;}
      break;
    case 99:
      // delay before invoking MainNext task, MainCnt = 50 for 1 second delay
      MainCnt--; if (MainCnt < 1) {MainTask = MainNext;}
      break;
    default:
      // error - task not defined so clear the pointer
      setMainTask(0); break;
  }
}

// ---------------------------------------------------------------------------------

void Task_AI() {
  // this is the main task for all bot movements including demo functions
  // it is primed by other tasks which set conditions for the AI to follow
  SpdChg = false;  // flag tracking wheel speed changes
  IR_Mask2 = IR_Mask & B11000;  // create a 2-bit mask
  IR_Mask3 = IR_Mask & B00111;  // create a 3-bit mask
  if (WiFiConnected) {PrintTgt = 1;}  // direct data to the WiFi link if connected
  switch(AI_Sub) {
    case 0: // first entry point for AI
      if (CountDown > 0) {
        //#########################################################
        // COUNT DOWN tasks
        //#########################################################
        if (AI_REPORT) {PrintTx+= "CountDown " + String(CountDown) + "\n";}
        Display_Count(CountDown);
        LED_Task = -60; // display the CountDown value in LEDs
        if (CountDown == 3){
          if (AI_LTOF_ON) {
            // turn ON range finder scanner?
            ServoTaskPnt = -2; ServoAuto = true; VL53L0X_ON ();
            if (AI_REPORT) {PrintTx+= "LTOF & servo are ON\n";}
          } else {ServoTaskPnt = -1; VL53L0X_OFF ();}
        }
        if (CountDown == 1){
          // turn ON IR fast scanning?
          if (AI_IR_ON) {
            ADC_Fast = true;  // start high speed sensor scanning
            initialiseIR();   // determine surface mode, white or black
            if (AI_REPORT) {PrintTx+= "IR sensors are ON\n";}
          }
          AI_Sub++;
        }
        CountDown--; MainDel = 50; // set skip timer and exit
      } else {
        // no countdown specified in this run
        if (AI_IR_ON) {
          ADC_Fast = true;  // start high speed sensor scanning
          initialiseIR();   // determine surface mode, white or black
          if (AI_REPORT) {PrintTx+= "IR sensors are ON\n";}
        }
        if (AI_LTOF_ON) {
          // turn ON range finder scanner?
          ServoTaskPnt = -2; ServoAuto = true; VL53L0X_ON ();
          if (AI_REPORT) {PrintTx+= "LTOF & servo are ON\n";}
        } else {ServoTaskPnt = -1; VL53L0X_OFF ();}
        if (WiFiConnected) {PrintTgt = 1;}  // direct data to the WiFi link if connected
        AI_Sub++;
      }
      break;
    case 1: // start
      //#########################################################
      // START task
      //#########################################################
      Display_Title24(AI_Title$);       // display this whilst moving and lock display
      L_Spd = 0; R_Spd = 0;             // reset wheel speed variables
      ClearDistance();                  // clear distance counters
      LED_Task = AI_LED_Task;           // LEDs mapped to IR_Mask
      Turn_Flg = 0;                     // cancel any previous turns
      IR_Mask3 = 0;                     // clear the IR sensor mask
      RndDel = 0; RndTask = -1;         // initialise random drive tasks
      Attack = false;                   // clear the attack flag
      if (RangeFtdTgt >= RangeTgtAtt) {
        // normal start
        setWheelSpeed(SpeedSrt,SpeedSrt); // off we go! This sets L_Spd and R_Spd
        Speed = SpeedSrt;                 // load the start speed
      } else {
        // facing oponent so random start, we will either attack or perform a tactical
        // movement
        switch (random(1)) {
          case 0: // attack
            break;
          case 1: // tactical move
            break;
        }
      }
      AI_Sub = 10; break;

    case 10:  // main loop start point
      //#########################################################
      // SURVIVAL tasks
      //#########################################################
      if (ItsOver) {setMainTask(0); return;}  // Oh dear, it's over!
      // perform edge detection checks
      if (AI_EDGE_ON) {if (Task_Edge_Check() > 0) {return;}}
      if (AI_TIP_ON) {
        if ((Tip > 0) || (TipLast > 0)) {Tip_Response();}
      }
      //#########################################################
      // START speed ramp up
      //#########################################################
      // ensure that we reach the target speed irrespective of other events
      if (Speed < SpeedTgt) {
        Speed += SpeedInc; L_Spd = Speed; R_Spd = Speed; SpdChg = true;
        if (Speed >= SpeedTgt) {
          Speed = SpeedTgt;
          if (AI_REPORT) {PrintTx+= "Speed at SpeedTgt " + String(SpeedTgt) + "\n";}
        }
      }
      if (!Attack) {
        //#########################################################
        // HUNTING & DRIVING tasks
        //#########################################################
        // not trying to survive anything so look for target if LTOF is ON
        Drive_En = true;  // enables random driving function, unless cancelled by steering
        if (AI_LTOF_ON) {
          if (LED_Task != 30) {LED_Task = -30;}
          if (RangeFtdTgt < RangeTgtUL) { 
            //#########################################################
            // HUNTING tasks
            //#########################################################
            // a target is being detected so adjust steering towards it
            // the steering adjustment is inversely based on range
            Drive_En = false;     // cancel random steering
            RndTask = -1;         // reset random driving to straight initially
            Task_Target_Steer();  // steer towards target
          }
          if (RangeFtdTgt <= RangeTgtAtt) {
            // a target is within attacking range, but are we pointing at it?
            if ((ServoAngTgtAvg >= 75) && (ServoAngTgtAvg <= 105)) {
              //#########################################################
              // Trigger ATTACK tasks
              //#########################################################
              // Yes, we are looking at it
              if (AI_ATTACK_EN) {
                // set the Attack flag, if this mode is enabled
                Drive_En = false; // disable random steering when entering ATTACK mode
                L_Spd = SpeedMax; R_Spd = SpeedMax; SpdChg = true;
                Steer = 0; Attack = true; LED_Task = -50; AI_ATT_CNT = 0;
                if (AI_REPORT) {
                  PrintTx+= "Switching to ATTACK\n";
                  PrintTx+= "LTOF range " + String(RangeFtdTgt) + "\n";
                }
              }
            }
          }
        }
        if (Drive_En) {
          //#########################################################
          // DRIVING tasks
          //#########################################################
          // not targetting, so apply simple turns at random intervals
          Task_Random_Drive();
        }          
     } else {
        //#########################################################
        // ATTACKING tasks
        //#########################################################
        // continue attacking and steer towards the target
        // if the target disappears then go straight ahead after a delay
        if (RangeFtdTgt > 300) {
          //#########################################################
          // TARGET LOST? tasks
          //#########################################################
          // target has moved away, so cancel attack after a delay
          AI_ATT_CNT++;
          if (AI_ATT_CNT >= 50) {
            L_Spd = SpeedMax; R_Spd = SpeedMax; // drop speed
            if (AI_REPORT) {
              PrintTx+= "Target lost at " + String(RangeFtdTgt) + "mm\n";
            }
          } else {Task_Target_Steer();}
        } else {
          //#########################################################
          // ATTACKING continues
          //#########################################################
          // adjust steering towards target
          Task_Target_Steer();
          AI_ATT_CNT = 0;
        }
      }
      break;

    case 99:
      // used to exit AI mode from any function after a delay
      MainCnt--; if (MainCnt < 1) {setMainTask(0);}
      break;
  }
  // if a wheel speed change has occured then set the new speed
  if (SpdChg) {
    // ensure that wheel speeds are within limits
    L_Spd = constrain(L_Spd,-1000,1000); R_Spd = constrain(R_Spd,-1000,1000);
    setWheelSpeed(L_Spd,R_Spd);
    if (AI_REPORT) {
      PrintTx+= "L_Spd " + String(L_Spd) + "\tR_Spd " + String(R_Spd) + "\n";
//      PrintTx+= "L_Dst " + String(L_Dst/100) + "\tR_Dst " + String(R_Dst/100) + "\n";
    }
  }
}

// ---------------------------------------------------------------------------------

void Task_Cal_IR() {
  // calibrate IR sensors
  switch(MainSub) {
    case 0: // pause after button press to calibration display message
      // revert limits to original measured values
      AvTmp[0] = (AvBlk[0] - AvWht[0])/8; AvWht[0]-= AvTmp[0]; AvBlk[0]+= AvTmp[0];
      AvTmp[1] = (AvBlk[1] - AvWht[1])/8; AvWht[1]-= AvTmp[1]; AvBlk[1]+= AvTmp[1];
      AvTmp[2] = (AvBlk[2] - AvWht[2])/8; AvWht[2]-= AvTmp[2]; AvBlk[2]+= AvTmp[2];
      AvTmp[3] = (AvBlk[3] - AvWht[3])/8; AvWht[3]-= AvTmp[3]; AvBlk[3]+= AvTmp[3];
      AvTmp[4] = (AvBlk[4] - AvWht[4])/8; AvWht[4]-= AvTmp[4]; AvBlk[4]+= AvTmp[4];
      MainSub++; MainDel = 50; break;
    case 1: // place Sumo Bot on white card
      Display_QText2x16("Stand me on a","WHITE surface");
      initialiseIR(); LED_Task = -20;
      MainSub++; ADC_Fast = true; break;
    case 2: // check for white card button press
      if (sw0State == LOW) {MainSub = 6; MainDel = 50; Display_Title24("< NO >");} // <NO button pressed
      if (sw1State == LOW) {MainSub++; MainDel = 50; Display_Title24("< OK >");}   // OK> button pressed
      break;
    case 3: // start calibrating white
      Display_Text2x16("Calibrating","WHITE surface");
      AvTmp[0] = 0; AvTmp[1] = 0; AvTmp[2] = 0; AvTmp[3] = 0; AvTmp[4] = 0;
      LED_Task = -40;
      MainSub++; MainCnt = 20; MainDel = 100; ADCscan = false; break;
    case 4: // calibrate by taking the average of 20 readings
      if (ADCscan) {
        AvTmp[0] += AvRaw[0]; AvTmp[1] += AvRaw[1]; AvTmp[2] += AvRaw[2];
        AvTmp[3] += AvRaw[3]; AvTmp[4] += AvRaw[4]; ADCscan = false;
        MainCnt--; if (MainCnt< 1) {MainSub++;}
      } break;
    case 5: // calc averages and update limits
      AvTmp[0]/= 20; AvTmp[1]/= 20; AvTmp[2]/= 20; AvTmp[3]/= 20; AvTmp[4]/= 20;
      AvWht[0] = AvTmp[0]; AvWht[1] = AvTmp[1]; AvWht[2] = AvTmp[2]; AvWht[3] = AvTmp[3]; AvWht[4] = AvTmp[4];
      MainSub++; MainDel = 100; break;
    case 6:
      Display_QText2x16("Stand me on a","BLACK surface");
      LED_Task = -20; MainSub++; WaitUp = true; break;
    case 7: // check for black card button press
      // note that moving through the air will be black
      if (sw0State == LOW) {MainSub = 11; MainDel = 50; Display_Title24("< NO >");} // <NO button pressed
      if (sw1State == LOW) {MainSub++; MainDel = 50; Display_Title24("< OK >");}   // OK> button pressed
      break;
    case 8: // start calibrating white
      Display_Text2x16("Calibrating","BLACK surface");
      AvTmp[0] = 0; AvTmp[1] = 0; AvTmp[2] = 0; AvTmp[3] = 0; AvTmp[4] = 0;
      LED_Task = -40;
      MainSub++; MainCnt = 20; MainDel = 100;  ADCscan = false; break;
    case 9: // calibrate by taking the average of 20 readings
      if (ADCscan) {
        AvTmp[0] += AvRaw[0]; AvTmp[1] += AvRaw[1]; AvTmp[2] += AvRaw[2];
        AvTmp[3] += AvRaw[3]; AvTmp[4] += AvRaw[4]; ADCscan = false;
        MainCnt--; if (MainCnt< 1) {MainSub++;}
      } break;
    case 10: // calc averages and set threshold values
      AvTmp[0]/= 20; AvTmp[1]/= 20; AvTmp[2]/= 20; AvTmp[3]/= 20; AvTmp[4]/= 20;
      AvBlk[0] = AvTmp[0]; AvBlk[1] = AvTmp[1]; AvBlk[2] = AvTmp[2]; AvBlk[3] = AvTmp[3]; AvBlk[4] = AvTmp[4];
      MainSub++; MainDel = 100; break;
    case 11:
      // determine thresholds, set 10% in from measured limits
      AvTmp[0] = (AvBlk[0] - AvWht[0])/10; AvWht[0]+= AvTmp[0]; AvBlk[0]-= AvTmp[0];
      AvTmp[1] = (AvBlk[1] - AvWht[1])/10; AvWht[1]+= AvTmp[1]; AvBlk[1]-= AvTmp[1];
      AvTmp[2] = (AvBlk[2] - AvWht[2])/10; AvWht[2]+= AvTmp[2]; AvBlk[2]-= AvTmp[2];
      AvTmp[3] = (AvBlk[3] - AvWht[3])/10; AvWht[3]+= AvTmp[3]; AvBlk[3]-= AvTmp[3];
      AvTmp[4] = (AvBlk[4] - AvWht[4])/10; AvWht[4]+= AvTmp[4]; AvBlk[4]-= AvTmp[4];
      // now determine whether to save values?
      Display_QText2x16("Save values to","Flash memory?");
      LED_Task = -20; MainSub++; WaitUp = true; break;
    case 12: // user selection <NO> or <YES>?
      if (sw0State == LOW) {MainSub = 14; MainDel = 50; Display_Title24("< NO >");} // goto exit
      if (sw1State == LOW) {MainSub++; MainDel = 50; Display_Title24("< OK >");} // goto save
      break;
    case 13:  // saving limits
      Display_Text2x16("Values saved","...");
      EEPROM_Write();
      MainSub++; MainDel = 100; // set delay
      break;
    case 14:  // exit
      reportCalData();  // report values, whether updated or not
      setMainTask(98); MainNext = 0;  // ensure switches are released
      ADC_Fast = true;  // maintain ADC fast scan
      SW0_Nop = false; SW1_Nop = false; // re-enable switches
      LED_Task = -20; // we are still in IR sensor more so set LED mode
      break;
  }
}

// ---------------------------------------------------------------------------------

void Task_Cal_LTOF() {
  // calibrate LTOF servo
  // if a button is pressed at this point then exit
  if ((MainSub < 11) && ((sw0State == LOW) || (sw1State == LOW))) {MainSub = 13;}  // EXIT
  
  switch(MainSub) {
    case 0: // pause after button press to display message
      LED_Last = -abs(LED_Task);
      MainSub++; MainDel = 50; break;
    case 1: // place Sumo Bot on its back
      Display_Text2x16("Place me on", "my back");
      MainSub++; MainDel = 100; MainCnt = 100; break;
    case 2: // test for being on back > 60 Deg.
      if (PitchGyr > 60) {MainSub++; MainDel = 100;}
      break;
    case 3: // start calibrating LTOF servo
      // move to centre position and ensure full range is possible
      Display_Text2x16("Calibrating","LTOF servo");
      MainInt = LTOF_On; if (LTOF_On < 1) {VL53L0X_ON();}
      ServoTaskPnt = 0; 
      if (!ServoEn) {ServoON();}
      servoVal = ServoCtr; microServo.writeMicroseconds(servoVal);
      ServoAng = 90; LED_Task = -30;
      MainSub++; MainDel = 50; break; // allow it time to get there
    case 4: // check range at centre position
      if (RangeFtd == 600) {MainSub++;}
      else {MainSub = 90; Error$ = "Short range";}  // error condition
      break;
    case 5: // prep to find max PWM value
      DispNow = 1; DispNOP = false; // turn ON special display mode
      RangeFtdNew = false; MainCnt = 0; MainSub++; break;
    case 6: // find max PWM value, at which point range will drop off
      if (RangeFtd < 200) {
        // starting to see the edge
        ServoMaxNew = servoVal; MainSub++; MainDel = 100;
      } else {
        if (RangeFtdNew) {
          // LTOF has got a new range
          RangeFtdNew = false; MainCnt = 0;
          servoVal+= 1; microServo.writeMicroseconds(servoVal);
//        Serial.println(servoVal);
          ServoAng = getServoAng();  // get angle based on current limits
        } else {
          MainCnt++;
          if (MainCnt > 5) {MainSub = 90; Error$ = "No range";} // LTOF error, no range returned
        }
        if (servoVal >= ServoMaxDef) {MainSub = 90; Error$ = "No edge";} // error, edge not found
//        MainDel = 1;  // run at 1/2 speed
      } break;
    case 7: // prep find min PWM value from centre
      servoVal = ServoCtr; microServo.writeMicroseconds(servoVal);
      ServoAng = 90;
      MainSub++; MainDel = 150; break; // allow it time to get there
    case 8: // wait for range to reset to max
      microServo.writeMicroseconds(servoVal);
      if (RangeFtd >= 600) {RangeFtdNew = false; MainCnt = 0; MainSub++;}  // range has reset
      break;
    case 9: // find min PWM value from centre
      if (RangeFtd < 200) {
        // starting to see the edge
        ServoMinNew = servoVal; MainSub++; MainDel = 100;
      } else {
        if (RangeFtdNew) {
          // LTOF has got a new range
          RangeFtdNew = false; MainCnt = 0;
          servoVal-= 1; microServo.writeMicroseconds(servoVal);
//        Serial.println(servoVal);
          ServoAng = getServoAng();  // get angle based on current limits
        } else {
          MainCnt++;
          if (MainCnt > 5) {MainSub = 90; Error$ = "No range";} // LTOF error, no range returned
        }
        if (servoVal <= ServoMinDef) {MainSub = 90; Error$ = "No edge";} // error, edge not found
//        MainDel = 1;  // run at 1/2 speed
      } break;
    case 10: // return to centre and ask save Yes/No
      Serial.println("ServoMin " + String(ServoMinNew) + "\tServoMax " + String(ServoMaxNew));
      servoVal = ServoCtr; microServo.writeMicroseconds(servoVal);
      ServoAng = 90; DispNow = 0; // return display mode to normal
      Display_QText2x16("Save values to","Flash memory?");
      LED_Task = -1; MainSub++; break;
    case 11: // user selection <NO> or <OK>?
      if (sw0State == LOW) {MainSub = 13; MainDel = 50; Display_Title24("< NO >");} // <NO button pressed
      if (sw1State == LOW) {MainSub++; MainDel = 50; Display_Title24("< OK >");}   // OK> button pressed
      break;
    case 12:  // saving limits
      // add a small PWM margin of 50us to measurements
      Display_Text2x16("Values saved","...");
      ServoMin = ServoMinNew + 50; ServoMax = ServoMaxNew - 50; // update values
      ServoCtr = ServoMin + ((ServoMax - ServoMin)/2);
      EEPROM_Write();
      MainSub++; MainDel = 100; // set delay
      break;
    case 13: // exit point
      if (MainInt < 1) {VL53L0X_OFF();} // turn OFF laser if it wasn't ON to start with
      reportCalData();                  // report values, whether updated or not
      setMainTask(98); MainNext = 0;    // ensure switches are released
      SW0_Nop = false; SW1_Nop = false; // re-enable switches
      LED_Task = LED_Last;              // return to previous LED task
      break;

    case 90:  // error condition
      Display_Text2x16("ERROR!", Error$);
      MainSub = 13; MainDel = 100; break;  // exit after a 2s delay
  }
}

// ---------------------------------------------------------------------------------

void Task_Cal_MPU() {
  // calibrate MPU offsets
  switch(MainSub) {
    case 0: // pause after button press to display message
      MainSub++; MainDel = 50; break;
    case 1: // place Sumo Bot on white card
      Display_QText2x16("Stand me on a","LEVEL surface");
      MainSub++; break;
    case 2: // check for white card button press
      if (sw0State == LOW) {MainSub = 8; MainDel = 50; Display_Title24("< NO >");} // <NO button pressed
      if (sw1State == LOW) {MainSub++; MainDel = 50; Display_Title24("< OK >");}   // OK> button pressed
      break;
    case 3: // start calibrating
      Display_Text2x16("Calibrating","MPU sensors");
      MPUCal = true; AcXOff = 0; AcYOff = 0; AcZOff = 0;// zero current offsets
      AcXSum = 0; AcYSum = 0; AcZSum= 0; AcAvgDiv = 0;  // reset averaging
      GyXOff = 0; GyYOff = 0; GyZOff = 0; GyCal = 250;  // zero current offsets
      GyXSum = 0; GyYSum = 0; GyZSum= 0; GyAvgDiv = 0;  // reset averaging
      LED_Task = -40; MainSub++; MainDel = 200; break;
    case 4: // update offsets after 5 sec delay
      AcXOff = -AcXAvg; // set yaw offset error
      AcYOff = 0;       // value determined from robot being laid on its side
      AcZOff = -AcZAvg; // set pitch offset error
      MainSub++; MainDel = 100; break;
    case 5:
      MPUCal = false;  // force Gyro offset calibration
      Display_QText2x16("Save values to","Flash memory?");
      LED_Task = -1; MainSub++; break;
    case 6: // user selection <NO> or <YES>?
      if (sw0State == LOW) {MainSub = 8; MainDel = 50; Display_Title24("< NO >");} // <NO button pressed
      if (sw1State == LOW) {MainSub++; MainDel = 50; Display_Title24("< OK >");}   // OK> button pressed
      break;
    case 7:  // saving limits
      Display_Text2x16("Values saved","...");
      EEPROM_Write();
      MainSub++; MainDel = 100; // set delay
      break;
    case 8:  // exit
      reportCalData();                // report values, whether updated or not
      setMainTask(98); MainNext = 0;  // ensure switches are released
      SW0_Nop = false; SW1_Nop = false; // re-enable switches
      break;
  }
}

// ---------------------------------------------------------------------------------

void Task_Distance() {
  // demonstrates the use of the distance integrators by travelling a pre-defined
  // distance of 250mm, on a sheet of A4 paper
  switch(MainSub) {
    case 0: // initialise task
      Display_Title24("DISTANCE");
      ClearDistance();                    // clear distance counters
      if (WiFiConnected) {PrintTgt = 1;}  // direct data to the WiFi link if connected
      if (AI_REPORT) {PrintTx += "\nDistance Test 250mm...\n";}
      WaitUp = true; MainSub++; break;
    case 1: // set the initial speed
      L_Spd = SpeedMin; R_Spd = SpeedMin; // off we go! This sets L_Spd and R_Spd
      SpdChg = true; MainSub++; break;
    case 2: // increase speed to max
      if (L_Spd < SpeedMax) {L_Spd+= 50; R_Spd = L_Spd; SpdChg = true;}
      else {MainSub++;}
      break;
    case 3: // stop when at target distance of 250mm
      updateDistance();
      if (L_Dst >= 2500000) {
        MoveSTOP(1); setMainTask(0);
        if (AI_REPORT) {PrintTx+= "Travelled 250mm\n";}
      } break;
  }
  if (SpdChg) {
    // ensure that wheel speeds are within limits
    L_Spd = constrain(L_Spd,-1000,1000); R_Spd = constrain(R_Spd,-1000,1000);
    setWheelSpeed(L_Spd,R_Spd);
//    if (AI_REPORT) {
//      PrintTx+= "L_Spd " + String(L_Spd) + "\tR_Spd " + String(R_Spd) + "\n";
//      PrintTx+= "L_Dst " + String(L_Dst) + "\tR_Dst " + String(R_Dst) + "\n";
//    }
    SpdChg = false;
  }
}

// ---------------------------------------------------------------------------------

void Task_Driving() {
  // AI will drive the sumo bot randomly within the confines of the ring
  // process starts after a 3 second count down
  CountDown = 3;          // start with a count down
  AI_EDGE_ON = true;      // enable edge detection
  AI_IR_ON = true;        // turn ON IR sensors
  AI_LED_Task = -20;      // set LEDs to display IR sensors
  AI_Title$ = "DRIVING";  // logo during AI tasks
  SpeedSrt = SpeedMin;    // initial speed
  SpeedTgt = 600;         // target max speed
  SpeedInc = 10;          // set thte rate at which Speed will be increased
  PrintTx += "\nDriving mode...\n";
  MainTask = 1;           // initiate AI mode
}

// ---------------------------------------------------------------------------------

byte Task_Edge_Check() {
  // called from a AI to perform edge detection
  // if Turn_Flg > 0 indicates action is already being taken
  if ((IR_Mask & B00011111) == B00011111) {
    //#########################################################
    // EMERGENCY STOP
    //#########################################################
    // EMERGENCY STOP assume Sumo Bot has been picked up or run off the ring
    MoveSTOP(1); if (AI_REPORT) {PrintTx += "EMERGENCY STOP!\n";}
    MainCnt = 50; AI_Sub = 99; return 255;
  }
  // correct IR_Mask for tipping
  if (TipLast == 0) {IR_Mask2 = 0; IR_Mask &= B00111;} // not tipping so only use front sensors
  else {IR_Mask3 = 0; IR_Mask &= B11000;}              // tipping so only use rear sensors
  
  if ((IR_Mask > 0) && (Turn_Flg == 0)) {
    //#########################################################
    // EDGE DETECTED
    //#########################################################
    // border detected so react to that
    // IR_Mask: xx,xx,xx,RL,RR,FR,CT,FL bit map
    Turn_Flg = IR_Mask; Turn_Cnt = 0; Turn_Time = 0;
    if (LED_Task != 20) {LED_Task = -20;}
    if (AI_REPORT) {PrintTx+= "IR Sensors " + getMaskStr() + "\n";}
    // Prioritise front sensors, provided we are not tipping
    if (IR_Mask3 > 0) {
      switch(IR_Mask3) {
        case B00000001: // FL only
          if (AI_REPORT) {PrintTx += "FL_Sen turning right\n";}
          L_Spd = 300; L_SpdTgt = 500; R_Spd = -150; R_SpdTgt = -500;
          RotTgt = 80 + random(90); Turning = 1;
          break;
        case B00000010: // Ctr only
          if (AI_REPORT) {PrintTx += "CTR_Sen random turn";}
          if (random(10) <= 4) {L_Spd = 300; L_SpdTgt = 600; R_Spd = -300; R_SpdTgt = -600; Turning = 1; PrintTx += "right\n";}
          else {L_Spd = -300; L_SpdTgt = -600; R_Spd = 300; R_SpdTgt = 600; Turning = -1; PrintTx += "left\n";}
          RotTgt = 140 + random(40);
          break;
        case B00000011: // CTR & FL
          if (AI_REPORT) {PrintTx += "CTR+FL_Sen turning right\n";}
          L_Spd = 300; L_SpdTgt = 600; R_Spd = -300; R_SpdTgt = -600; 
          RotTgt = 140 + random(40); Turning = 1;
          break;
        case B00000100: // FR only
          if (AI_REPORT) {PrintTx += "FR_Sen turning left\n";}
          L_Spd = -150; L_SpdTgt = -500; R_Spd = 300; R_SpdTgt = 500; 
          RotTgt = 80 + random(90); Turning = -1;
          break;
        case B00000101: // FL & FR
          if (AI_REPORT) {PrintTx += "FL+FR_Sen priority turn\n";}
          if (Av100[0] > Av100[2]) {L_Spd = 400; L_SpdTgt = 600; R_Spd = -400; R_SpdTgt = -600; Turning = 1;}
          else {L_Spd = -400; L_SpdTgt = -600; R_Spd = 400; R_SpdTgt = 600; Turning = -1;}
          RotTgt = 140 + random(40);
          break;
        case B00000110: // CTR & FR
          if (AI_REPORT) {PrintTx += "CTR+FR_Sen turning left\n";}
          L_Spd = -300; L_SpdTgt = -600; R_Spd = 300; R_SpdTgt = 600; 
          RotTgt = 140 + random(40); Turning = -1;
          break;
        case B00000111: // Ctr & FL & FR
          if (AI_REPORT) {PrintTx += "FL+CTR+FR_Sen random turning\n";}
          if (random(10) <= 4) {L_Spd = 400; L_SpdTgt = 600; R_Spd = -400; R_SpdTgt = -600; Turning = 1;}
          else {L_Spd = -400; L_SpdTgt = -600; R_Spd = 400; R_SpdTgt = 600; Turning = -1;}
          RotTgt = 140 + random(40);
          break;
      }
    }
    if (IR_Mask2 > 0) {
      switch(IR_Mask2) {
        case B00001000: // RR only
          if (AI_REPORT) {PrintTx += "RR_Sen turning right\n";}
          L_Spd = 300; L_SpdTgt = 500; R_Spd = -150; R_SpdTgt = -500;
          RotTgt = 80 + random(90); Turning = 1;
          break;
        case B00010000: // RL only
          if (AI_REPORT) {PrintTx += "RL_Sen turning left\n";}
          L_Spd = -150; L_SpdTgt = -500; R_Spd = 300; R_SpdTgt = 500; 
          RotTgt = 80 + random(90); Turning = -1;
          break;
        case B00011000: // RR & RL
          if (AI_REPORT) {PrintTx += "RR+RL_Sen random turning\n";}
          if (random(10) <= 4) {L_Spd = 400; L_SpdTgt = 600; R_Spd = -400; R_SpdTgt = -600; Turning = 1;}
          else {L_Spd = -400; L_SpdTgt = -600; R_Spd = 400; R_SpdTgt = 600; Turning = -1;}
          RotTgt = 140 + random(40);
          break;
      }
    }
    //#########################################################
    // START TURNING
    //#########################################################
    // setting turning in motion
    setWheelSpeed(L_Spd,R_Spd); RotGyr = 0.0; Steer = 0;
    if (AI_REPORT) {
      if (Attack) {PrintTx += "Cancelling ATTACK\n";}
      PrintTx += "Rotate Tgt " + String(RotTgt);
    } Attack = false;
    // centralise scanner if enabled?
    if (AI_LTOF_ON) {
      ServoSubTask = 3; // centralise servo scanner
      ServoTaskDel = 0; // remove any skip delay
    }
    
  } else if (Turn_Flg > 0) {
    //#########################################################
    // TURNING
    //#########################################################
    // in the process of turning from an edge
    // exit when either turn is complete or if half way and a target appears
    RotESC = false;
    if (abs(RotGyrInt) >= RotTgt) {RotESC = true;}
    if (AI_LTOF_ON) {
      if (RangeFtdTgt < RangeTgtUL) {
        // a target has been detected whilst turning, if we have completed at
        // least half the turn then exit
        if (abs(RotGyrInt) >= RotTgt/2) {RotESC = true;}
      }
    }
    if (RotESC) {
      //#########################################################
      // TURN COMPLETE
      //#########################################################
      // completed the turn so set speeds back to going forward
      if (AI_REPORT) {
        PrintTx += "\tAngle " + String(RotGyr) + " Deg.\n";
        PrintTx += "Edge turn end\n\n";
      }
      L_Spd = max(int(L_Spd),int(R_Spd)); R_Spd = L_Spd;
      SpdChg = true; Turn_Flg = 0; RndTask = -1;
    } else {
      //#########################################################
      // STILL TURNING
      //#########################################################
      // still turning, so adjust speed towards targets
      if (L_Spd != L_SpdTgt) {
             if (L_Spd < L_SpdTgt) {L_Spd+= 10; L_Spd = min(L_Spd,L_SpdTgt); SpdChg = true;}
        else if (L_Spd > L_SpdTgt) {L_Spd-= 10; L_Spd = max(L_SpdTgt,L_Spd); SpdChg = true;}
      }
      if (R_Spd != R_SpdTgt) {
             if (R_Spd < R_SpdTgt) {R_Spd+= 10; R_Spd = min(R_Spd,R_SpdTgt); SpdChg = true;}
        else if (R_Spd > R_SpdTgt) {R_Spd-= 10; R_Spd = max(R_SpdTgt,R_Spd); SpdChg = true;}
      }
      if (SpdChg) {setWheelSpeed(L_Spd,R_Spd);}
    }
  }
  return Turn_Flg;
}

// ---------------------------------------------------------------------------------

void Task_Hunting() {
  // AI will drive the sumo bot randomly within the confines of the ring
  // process starts after a 3 second count down
  // It steers towards objects but does not attack them
  CountDown = 3;        // start with a count down
  AI_EDGE_ON = true;    // enable edge detection
  AI_IR_ON = true;      // turn ON IR sensors
  AI_LED_Task = -20;    // set LEDs to display IR sensors
  AI_LTOF_ON = true;    // enable LTOF servo scanner
  AI_Title$ = "HUNTING";  // logo during AI tasks
  SpeedSrt = SpeedMin;
  SpeedTgt = 600;
  SpeedInc = 20;          // set thte rate at which Speed will be increased
  PrintTx += "\nHunting mode...\n";
  MainTask = 1;         // initiate AI mode
}

// ---------------------------------------------------------------------------------

void Task_Random_Drive() {
  // called from AI Task to perform random steering tasks
  if (RndTask < 0) {
    // -ve tasks reset pointers
    RndTask = -RndTask; Steer = 0;
  }
  switch (RndTask) {
    case 1:
      // wait for target speed to be achieve before going random
      // go straight initially for a short distance
      if (Speed == SpeedTgt) {RndTask++; RndAhead = 100;} // go straight 100mm
      break;
    case 2:
      // decide on the type of drive pattern to follow
      // if coming out of a tight turn then go straight for a distance
      RndTurn_Flg = 0;  // clear turning flag
      if (RndAhead > 0) {RndTask = 5;}
      else {
        if (random(8) < 6) {RndTask = 10;}
        else {RndTask = 20;}
      } break;
      
    case 5:
      //#########################################################
      // STRAIGHT AHEAD
      //#########################################################
      // this drives forward for a distance defined by RndAhead
      ClearDistance();
      if (AI_REPORT) {PrintTx+= "FORWARD only " + String(RndAhead) + "mm\n";}
      Steer = 0;  // remove any steering bias
      // the distance measure count is 10,000 times greater than the target
      RndAhead*= 10000; RndTask++;
      break;
    case 6:
      // return to random tasks once the distance has been covered
      updateDistance(); // keep updating the distance tracker whilst waiting
      // as we are going straight ahead we only need to consider one wheel
      if (L_Dst >= RndAhead) {
        if (AI_REPORT) {PrintTx+= "Travelled " + String(RndAhead/10000) + "mm\n";}
        RndTask = 2; RndAhead = 0;
      } break;
      
    case 10:
      //#########################################################
      // SIMPLY TURNING
      //#########################################################
      // this is a simple curved drive, veering left or right
      RndInc = 10 + random(40); // set speed change rate
      RndDel = 25 + random(50); // set timer <1 - 2 secs
      Steer = (Speed * (15 + random(11)))/100;  // bias between 15 - 25%
      if (AI_REPORT) {PrintTx+= "RANDOM steer " + String(Steer) + "\n";}
      if ( Turning != 0) {
        // if turning from an edge we bias the random steering that way
        if ( Turning > 0) {Steer = -Steer;}
        Turning = 0;
      } else {if (random(8) > 3) {Steer = -Steer;}} // 50/50 chance
      RndTask = 99; break;

    case 20:
      //#########################################################
      // NEUTRAL TURN
      //#########################################################
      // turn through a short random angle
      if (AI_LTOF_ON) {
        ServoSubTask = 3; // centralise servo scanner
        ServoTaskDel = 0; // remove any skip delay
        RndTurn_Flg = 1;  // set flag to hold off servo movement
      }
      RndInc = 50 + random(150); // set speed change rate
      Steer = (Speed * 8)/10; if (random(8) > 3) {Steer = -Steer;}
      if (AI_REPORT) {PrintTx+= "NEUTRAL steer " + String(Steer) + "\n";}
      RotTgt = 20 + random(250); // turn through 20 - 270 Deg.
      RotGyr = 0.0; RndTask++; break;
    case 21:
      // turn until angle is reached or a target appears
      if (AI_LTOF_ON && (RangeFtdTgt < RangeTgtUL)) {
        // just spotted a target
        Steer = 0; L_Spd = Speed; R_Spd = Speed; SpdChg = true;
        RndTurn_Flg = 0; RndAhead = 100; RndTask = 2;
      }
      if (abs(RotGyrInt) >= RotTgt) {
        // angle achieved, we need to stop turning
        // go straight initially before going random
        Steer = 0; RndInc = 50; RndTurn_Flg = 0; RndAhead = 100; RndTask = 2;
      } break;
      
    case 99:
      // wait for delay count to complete before returing to task 2
      RndDel--; if (RndDel < 1) {RndTask = 2;}
      break;
  }
  // adjust speed towards the speed + steer target
       if (L_Spd < (Speed + Steer)) {L_Spd+= RndInc; SpdChg = true; if (L_Spd > (Speed + Steer)) {L_Spd = (Speed + Steer);}}
  else if (L_Spd > (Speed + Steer)) {L_Spd-= RndInc; SpdChg = true; if (L_Spd < (Speed + Steer)) {L_Spd = (Speed + Steer);}}
       if (R_Spd < (Speed - Steer)) {R_Spd+= RndInc; SpdChg = true; if (R_Spd > (Speed - Steer)) {R_Spd = (Speed - Steer);}}
  else if (R_Spd > (Speed - Steer)) {R_Spd-= RndInc; SpdChg = true; if (R_Spd < (Speed - Steer)) {R_Spd = (Speed - Steer);}}
}

// ---------------------------------------------------------------------------------

void Task_SumoM0() {
  // AI will demonstrate Sumo 0 attacking mode 
  CountDown = 5;        // start with a count down
  AI_ATTACK_EN = true;  // if == true it enables AI attack mode
  AI_EDGE_ON = true;    // enable edge detection mode
  AI_IR_ON = true;      // turn ON IR sensors
  AI_LED_Task = -20;    // set LEDs to display IR sensors
  AI_LTOF_ON = true;    // enable LTOF servo scanner
  AI_TIP_ON = true;     // enable tip responses
  AI_Title$ = "SUMO M0";// logo during AI tasks
  SpeedInc = 20;        // auto-speed adjustment increment
  SpeedSrt = SpeedMin;  // starting speed
  SpeedTgt = 800;       // general target speed for auto-adjustment
  SpeedInc = 100;       // set thte rate at which Speed will be increased
  PrintTx += "\nSumo mode 0...\n";
  MainTask = 1;         // initiate AI mode
}

// ---------------------------------------------------------------------------------

void Task_Target() {
  // This task demonstrate target tracking using the AI method
  // Stop the bot by either pressing a button or picking it up 
  switch(MainSub) {
    case 0: // start the task
      Display_Title24("TARGET");
      if (WiFiConnected) {PrintTgt = 1;}  // direct data to the WiFi link if connected
      ServoTaskPnt = -2;  // start servo scanning
      ServoAuto = true;   // turn ON automatic centring
      LED_Task = -30;     // LEDs mapped to servo track
      VL53L0X_ON();       // turn LTOF laser ON
      ADC_Fast = true;    // start high speed sensor scanning
      initialiseIR();     // determine surface mode, white or black
      Speed = 0;          // start with no speed
      SpeedTgt = 800;     // target speed when moving from a distance
      MainCnt = 0; MainSub++; break;
    case 1: // run the task, exit if picked up
      SpdChg = false; // clear the speed change flag
      if (IR_Mask > 0) {
        // bot has left the ground so stop teh motors and exit
        MoveSTOP(0);    // stop the motors
        waitMainEnd();  // exit Main Task
        return;
      } else {
        // use range/angle to determine what happens
        if (RangeFtdTgt < RangeTgtAtt) {
          //#########################################################
          // ATTACKING task
          //#########################################################
          // a target is within attacking range, but are we pointing at it?
          if ((ServoAngTgtAvg >= 75) && (ServoAngTgtAvg <= 105)) {
            if (LED_Task != 50) {LED_Task = -50;}
            SpeedTgt = SpeedMax;
          }
          Task_Target_Steer();  // steer towards target
          if (SpdChg) {PrintTx+="A-";}
        } else if (RangeFtdTgt < RangeTgtUL) { 
          //#########################################################
          // HUNTING task
          //#########################################################
          // a target is being detected so adjust steering towards it
          // the steering adjustment is inversely based on range
          if (LED_Task != 30) {LED_Task = -30;}
          SpeedTgt = 800;       // hunt at reduced speed
          Task_Target_Steer();  // steer towards target
          if (SpdChg) {PrintTx+="H-";}
        } else if (RangeFtdTgt < 500) {
          //#########################################################
          // SENSE task
          //#########################################################
          // within target sensing so move off slowly
          if ((L_Spd < SpeedTgt) && (MainCnt < 1)) {
            L_Spd+= 50; L_Spd = min(L_Spd,SpeedTgt);
            if (L_Spd < SpeedMin) {L_Spd = SpeedMin;}
            R_Spd = L_Spd; SpdChg= true;
          } else {MainCnt = 1; Task_Target_Steer();}
          if (SpdChg) {PrintTx+="S-";}
        } else {
          //#########################################################
          // SLOW DOWN task
          //#########################################################
          // out of range so slow down if moving to a STOP
          L_Spd = max(L_Spd,R_Spd);
          if (L_Spd > SpeedMin) {L_Spd-= 50; SpdChg= true;}
          else if (L_Spd > 0) {L_Spd = 0; SpdChg= true;}
          R_Spd = L_Spd;
          if (SpdChg) {PrintTx+="!-";}
        }
      } break;
  }
  // if a wheel speed change has occured then set the new speed
  if (SpdChg) {
    // ensure that wheel speeds are within limits
    L_Spd = constrain(L_Spd,-1000,1000); R_Spd = constrain(R_Spd,-1000,1000);
    setWheelSpeed(L_Spd,R_Spd);
    if (AI_REPORT) {
      PrintTx+= "Rng " + String(RangeFtdTgt) + "\tAng " + String(ServoAngTgtAvg) + "\tL_Spd " + String(L_Spd) + "\tR_Spd " + String(R_Spd) + "\n";
//      PrintTx+= "L_Dst " + String(L_Dst/100) + "\tR_Dst " + String(R_Dst/100) + "\n";
    }
  }
}

// ---------------------------------------------------------------------------------

void Task_Target_Steer() {
  // called when target is within range, so steer towards target
  int16_t zSM = 90;                            // short range factor
  if (RangeFtdTgt > 150) {zSM = map(RangeFtdTgt,150,RangeTgtOR,90,20);} // range factor
  int16_t zSpd = ((90 - ServoAngTgtAvg) * zSM)/10; // speed delta
  int16_t zSD;                                // speed overshoot
  if (ServoAngTgtAvg < 88) {
    // target is to my left, zSpd will be +ve
    R_SpdTgt = SpeedTgt + zSpd; zSD = 0;  // increase right motor
    if (R_SpdTgt > SpeedMax) {zSD = R_SpdTgt - SpeedMax; R_SpdTgt = SpeedMax;}
    L_SpdTgt = SpeedTgt - zSpd - zSD;     // reduce left motor
  } else if (ServoAngTgtAvg > 92) {
    // target is to my right, zSpd will be -ve
    L_SpdTgt = SpeedTgt - zSpd; zSD = 0;  // increase left motor
    if (L_SpdTgt > SpeedMax) {zSD = L_SpdTgt - SpeedMax; L_SpdTgt = SpeedMax;}
    R_SpdTgt = SpeedTgt + zSpd - zSD;     // reduce right motor
  } else {
    // go straight ahead
    L_SpdTgt = SpeedTgt; R_SpdTgt = SpeedTgt;
  }
  // check for speed change
  if (L_Spd != L_SpdTgt) {L_Spd = L_SpdTgt; SpdChg = true;}
  if (R_Spd != R_SpdTgt) {R_Spd = R_SpdTgt; SpdChg = true;}
}

// ---------------------------------------------------------------------------------

void Task_TipTest() {
  // AI will demonstrate a responses to tipping
  CountDown = 0;        // start with a count down
  AI_EDGE_ON = true;    // enable edge detection mode
  AI_IR_ON = true;      // turn ON IR sensors
  AI_LED_Task = -20;    // set LEDs to display IR sensors
  AI_TIP_ON = true;     // enable tip responses
  AI_Title$ = "TIP TEST";// logo during AI tasks
  SpeedSrt = 0;         // starting speed
  SpeedTgt = 0;         // target speed
  PrintTx += "\nTip Test...\n";
  MainTask = 1;         // initiate AI mode
}

// ---------------------------------------------------------------------------------

void TestModeRun(int16_t zMode) {
  // invoke a special test mode
  switch(zMode) {
    case 0: // servo sweep rate test
      setMainTask(70); break;
  }
}

// ---------------------------------------------------------------------------------

void Test_SweepRate() {
  // runs a test sequence to determine the sweep rate of the micro-servo
  // it uses the sweep limit LTOF dark points to indicate when the servo has got there 
  // the test is repeated several times to give min,med and max delay times
  // its accuracy is limited by the LTOF rate of approx 33Hz
  // to avoid errors the test should be run with the robot on its back
  // test results are: min=320, med=351, max=385  Sumo 2
  // test results are: min=320, med=357, max=386  Sumo 1
  
  switch(MainSub) {
    case 0: // pause after button press to display message
      Display_Text2x16("Servo Sweep","Rate");
      if (LTOF_On < 1) {VL53L0X_ON();}
      ServoTaskPnt = 0; if (!ServoEn) {ServoON();}
      // set the two end servo values
      ServoMinNew = ServoMin - 50; ServoMaxNew = ServoMax + 50;
      servoVal = ServoMinNew; microServo.writeMicroseconds(servoVal);
      ServoAng = getServoAng(); LED_Last = LED_Task; LED_Task = -30;
      MainFor = 10; MainCnt = 5;  // set up the counters
      MainMax = 0; MainMin = 1000; MainSum = 0; // min/max and sum values
      MainSub++; MainDel = 50; break; // allow it time to get there
    case 1: // wait for the range to drop, and remain low
      if (RangeFtd < 600) {
        MainCnt--; if (MainCnt < 1) {MainSub++;}
      } else {MainCnt = 5;} // reset the glitch counter 
      break;
    case 2: // now start the sweep, recording the time of the latest Range measurement
      MainTime = LTOF_t1;
      servoVal = ServoMaxNew; microServo.writeMicroseconds(servoVal);
      ServoAng = getServoAng();
      MainCnt = 0; MainSub++; break;
    case 3: // range should increase whilst sweeping, or flag an error
      if (RangeFtd == 600) {
        Serial.print(String(LTOF_t1 - MainTime) + "\t");
        MainSub++;
      } else {MainCnt++; if (MainCnt > 50) {
        // during 1 second range has not increased to max, so this is an error
        Error$ = "Range Low"; MainSub= 90;
      }} break;
    case 4: // now look for short range at servo max PWM
      if (RangeFtd < 600) {
        MainTime = LTOF_t1 - MainTime; MainSub++;
      } else {MainCnt++; if (MainCnt > 50) {
        // during 1 second range has not decreased below 600, so this is an error
        Error$ = "Range Max"; MainSub = 90;
      }} break;
    case 5: // we have a result for min to max sweep
      MainMin = min(int16_t(MainTime),MainMin); MainMax = max(int16_t(MainTime),MainMax);
      MainSum+= MainTime;
      Serial.println("Servo min/max Rate " + String(MainTime) + "ms");
      Display_Text2x16("Servo Rate",String(MainTime) + "ms");
      MainSub++; MainDel = 50; break;
      
    case 6: // now sweep from max to min
      MainTime = LTOF_t1;
      servoVal = ServoMinNew; microServo.writeMicroseconds(servoVal);
      ServoAng = getServoAng();
      MainCnt = 0; MainSub++; break;
    case 7: // range should increase whilst sweeping, or flag an error
      if (RangeFtd == 600) {
        Serial.print(String(LTOF_t1 - MainTime) + "\t");
        MainSub++;
      } else {MainCnt++; if (MainCnt > 50) {
        // during 1 second range has not increased to max, so this is an error
        Error$ = "Range Low"; MainSub= 90;
      }} break;
    case 8: // now look for short range at servo min PWM
      if (RangeFtd < 600) {
        MainTime = LTOF_t1 - MainTime; MainSub++;
      } else {MainCnt++; if (MainCnt > 50) {
        // during 1 second range has not decreased below 600, so this is an error
        Error$ = "Range Min"; MainSub = 90;
      }} break;
    case 9: // we have a result for max to min sweep
      MainMin = min(int16_t(MainTime),MainMin); MainMax = max(int16_t(MainTime),MainMax);
      MainSum+= MainTime;
      Serial.println("Servo max/min Rate " + String(MainTime) + "ms");
      Display_Text2x16("Servo Rate",String(MainTime) + "ms");
      MainFor--; if (MainFor < 1) {MainSub = 10;} else {MainSub = 2;}
      MainDel = 50; break;
    case 10:
      MainSum/= 20; // get the average
      Serial.println("\n" + String(MainMin) + "\t" + String(MainSum) + "\t" +String(MainMax));
      MainSub++; break;

     case 11: // exit point
      setMainTask(98); MainNext = 0;    // ensure switches are released
      LED_Task = LED_Last;              // return to previous LED task
      break;

    case 90:  // error condition
      Display_Text2x16("ERROR!", Error$);
      MainSub = 11; MainDel = 100; break;  // exit after a 2s delay
  }
}

// ---------------------------------------------------------------------------------

void Tip_Response() {
  // called from AI when Tip > 0 to determine a response
  // 1 = pitch tipping backwards
  // 2 = yaw tipping right
  // 3 = yaw tipping left
  // The Tip flag is reset constantly when calculating angles
  int zLS = L_Spd;  // current speed
  int zRS = R_Spd;  // current speed
//  PrintTx += String(Tip) + "\t";
  if (TipLast != Tip) {
    // We have already received a previous Tip?, check to see if it has changed
    if (Tip > 0) {TipLast = 0;}  // changed so reset the historic flag
  }
  if (TipLast == 0) {
    // a new tip condition
    TipLast = Tip;
    if ((Tip & 1) == 1) {
      // Pitch gyro indicates ang. >= 5 Deg. tip
      RunCnt = 50;  // set over-run timer at 500ms
      if (PitchGyrInt < 40) {
        // tipping backwards up to 45 Deg.
        // drive backwards, fast response, slow decay
        // respond to rear sensors or random turn
        if ((IR_Mask & B00011000) == B00011000) {
          // both rear sensors are active so no where to go, Oh dear!
          L_Spd = 0; R_Spd = 0; SpdChg = true;
        } else if ((IR_Mask & B00010000) > 0) {
          // rear left sensor is active
          L_Spd = 0; R_Spd = -SpeedMax/2; SpdChg = true;
        } else if ((IR_Mask & B00001000) > 0) {
          // rear right sensor is active
          L_Spd = -SpeedMax/2; R_Spd = 0; SpdChg = true;
        } else {
          // not at edge so turn in a sweeping move
          if (TipRnd > 0) {L_Spd = -SpeedMax; R_Spd = -SpeedMax/2;
          } else {L_Spd = -SpeedMax/2; R_Spd = -SpeedMax;}
           SpdChg = true;
        }
//        PrintTx += "P" + String(PitchGyrInt) + "D\t" + String(L_Spd) + "\t" + String(R_Spd);
      }
    } else if ((Tip & 2) == 2) {
      // tipping sideways at Yaw >= 15 Deg
      // Yaw -ve tipping to the right, +ve tipping to the left
      // use the rotation gyro to decide which way to drive the active wheel
      if (YawGyrInt > 0) {
        // tipping to the left
        // RotGyr + if turning clockwise
        if (RotGyrInt > 0) {L_Spd = -SpeedMax; R_Spd = 0;}
        else {L_Spd = 0; R_Spd = SpeedMax;}
         SpdChg = true;
      } else {
        // tipping to the right
        // RotGyr + if turning clockwise
        if (RotGyrInt > 0) {L_Spd = 0; R_Spd = -SpeedMax;}
        else {L_Spd = 0; R_Spd = SpeedMax;}
        SpdChg = true;
      }
//      PrintTx += "Y" + String(YawGyrInt) + "D\t" + String(L_Spd) + "\t" + String(R_Spd);
    }
  } else if (Tip == 0) {
    // tipping has ceased, so turn off wheel motors if active
    // after a short over-run
    if (RunCnt > 0) {
      RunCnt--;
      if (RunCnt < 10) {
        // towards the end of over-run revert to normal speed
        if ((L_Spd != SpeedTgt) || (R_Spd != SpeedTgt)) {
          if (L_Spd < SpeedTgt) {L_Spd += 50;} else {L_Spd = SpeedTgt;}
          if (L_Spd > SpeedTgt) {L_Spd -= 50;} else {L_Spd = SpeedTgt;}
          if (R_Spd < SpeedTgt) {R_Spd += 50;} else {R_Spd = SpeedTgt;}
          if (R_Spd > SpeedTgt) {R_Spd -= 50;} else {R_Spd = SpeedTgt;}
          SpdChg = true;
        }
      }
    } else {
      // end/exit of Tip response routine
      L_Spd = SpeedTgt; R_Spd = SpeedTgt; SpdChg = true;
      TipRnd = random(2); TipLast = 0;
    }
//    PrintTx += String(RunCnt) + "\t" + String(L_Spd) + "\t" + String(R_Spd);
  } else {
    // still responding to Tip condition
    // wait for either no tipping or tipped over conditions
    if (abs(PitchGyrInt) > 45) {ItsOver = true;}
    if (abs(YawGyrInt) > 45) {ItsOver = true;}
    if ((IR_Mask & B00011000) == B00011000) {setWheelSpeed(0,0); TipLast = 0;}
    if (AI_REPORT) {if ((Tip & 2) > 0) {PrintTx += "Yaw Tip\n";} else {PrintTx += "Pitch Tip\n";}}
//    PrintTx += String(PitchGyrInt) + "D\t" + String(L_Spd) + "\t" + String(R_Spd);
  }
}

// ---------------------------------------------------------------------------------
