
/* SSD1306Wire.h functions, see https://github.com/ThingPulse/esp8266-oled-ssd1306

  disply.clear()        - clear the local pixel buffer
  disply.display()      - write the buffer to the display
  disply.displayOff()   - turn the display OFF
  disply.displayOn()    - turn the display ON
  display.drawCircle(int16_t x, int16_t y, int16_t radius)
  display.drawFastImage(int16_t x, int16_t y, int16_t width, int16_t height, const uint8_t *image)
  display.drawHorizontalLine(int16_t x, int16_t y, int16_t length)
  display.drawLine(int16_t x0, int16_t y0, int16_t x1, int16_t y1)  - draw from x0k,y0 to x1,y1
  display.drawProgressBar(uint16_t x, uint16_t y, uint16_t width, uint16_t height, uint8_t progress)
  display.drawRect(int16_t x, int16_t y, int16_t width, int16_t height)
  display.drawString(int16_t x, int16_t y, String text)
  display.drawStringMaxWidth(int16_t x, int16_t y, int16_t maxLineWidth, String text)
  display.drawVerticalLine(int16_t x, int16_t y, int16_t length)
  display.drawXbm(int16_t x, int16_t y, int16_t width, int16_t height, const char* xbm)
  display.fillCircle(int16_t x, int16_t y, int16_t radius)
  display.fillRect(int16_t x, int16_t y, int16_t width, int16_t height)
  display.flipScreenVertically()   - flip the current screen upside down
  disply.init()         - initialise the display
  disply.invertDisplay()- inverted display mode
  disply.end()          - free memory used by the display
  disply.displayOn()    - turn the display ON
  display.mirrorScreen() - draw the screen mirrored
  disply.normalDisplay()- normal display mode
  disply.reconnect()    - connect again through I2C
  display.setBrightness(uint8_t)
  display.setColor(OLEDDISPLAY_COLOR color) - BLACK, WHITE, INVERSE
  display.setContrast(uint8_t contrast, uint8_t precharge = 241, uint8_t comdetect = 64)
  display.setFont(const uint8_t* fontData)
  display.setPixel(int16_t x, int16_t y)  - set a pixel at x,y
  display.setTextAlignment(OLEDDISPLAY_TEXT_ALIGNMENT textAlignment)  - TEXT_ALIGN_LEFT, TEXT_ALIGN_CENTER, TEXT_ALIGN_RIGHT, TEXT_ALIGN_CENTER_BOTH

  W = display.getStringWidth(const char* text, uint16_t length)
  W = display.getStringWidth(String text)
  
  Display size is 128 x 64 (w x h) pixels
  Default fonts are ArialMT at 10, 16 & 24 pixels high
  Text options include:
    2 x 24, + 1 x 10 pnt lines
    1 x 24, + 2 x 16 pnt lines
    1 x 24, + 3 x 10 pnt lines
    4 x 16 pnt lines, no line spaces, decending characters will merge
    3 x 16, + 1 x 10 pnt lines
    6 x 10 pnt lines, no line spaces, decending characters will merge
*/

// --------------------------------------------------------------------------------

void Display_40ms() {
  // called every 40ms from the main loop
  // writes display data indirectly by calling DispMon functions, which allows a
  // remote Windows app to be used to also display the data
  if (DispNOP) {return;}  // this function is aborted, no I2C activity
  
//  if (DispDel > 0) {DispDel--; return;} // don't display if DispDel > 0
  if (DispClr) {Display_clear(); DispClr = false; return;}
  if (DispDisp) {Display_display(); DispDisp = false; DispClr = true; return;}

  // we refresh individual displays based on a counter, DispCnt
  // set DispCnt -ve to prevent the display update process
  DispCnt--;
  if (DispCnt < 1) {
    // DispCnt is reset for display updates, so can be different for each
    if (DispBack) {
      // draw a back arrow on the screen
      Display_clear();
      Display_drawLine(48,32,80,16);
      Display_drawLine(48,32,80,48);
      Display_drawLine(80,16,80,48);
      DispCnt = 10;
    } else {
      if (DispNow > 0) {
        // special display modes used in calibration etc.
        switch(DispNow) {
          case 1: DispCnt = 3; Display_ServoVals(); break;
        }
      } else {
        // normal menu screens, display depends on menu modes and AI activitiy
        switch(DispMenu) {
          case 0: Display_40ms_Sensors(); break;
          case 1: Display_40ms_Tasks(); break;
          case 2: Display_40ms_Sumo(); break;
        }
      }
    }
    DispDisp = true; // having created the display contents, set flag to display it
    // the following line puts a rectangle around every display, showing the boundry region
//    display.drawRect(0,0,128,64);
  }
}

// --------------------------------------------------------------------------------

void Display_40ms_Sensors() {
  // DispMenu == 0 so display sensor screens
  switch (DispPnt) {
    case 0: // Battery screens
      switch(DispMode) {
        case 0: DispCnt = 10; Display_Battery(); break;
        case 1: DispCnt = 10; Display_BatData(); break;
        default: DispMode = 0; break; // over-run, returns display to DispMode = 0
      } break;
    case 1: // MPU6050 screens
      switch(DispMode) {
        case 0: DispCnt = 3; Display_MPU_Levels(); break;
        case 1: DispCnt = 3; Display_MPU_Rot(); break;
        case 2: DispCnt = 5; Display_MPU_Acc(); break;
        case 3: DispCnt = 5; Display_MPU_Gyr(); break;
        case 4: DispCnt = 5; Display_MPU_Ang(); break;
        default: DispMode = 0; break; // over-run, returns display to DispMode = 0
      } break;
    case 2: // LTOF ranger screens
      switch(DispMode) {
        case 0: // display LTOF inactive
          if (DispChng) {
            if (LTOF_On > 0) {VL53L0X_OFF();}
            ServoTaskPnt = -1;  // centre servo
          } DispCnt = 10; Display_LTOF(); break;
        case 1: // display LTOF range values
          if (DispChng) {
            if (LTOF_On < 1) {VL53L0X_ON();}
            ServoTaskPnt = -1;  // centre servo
          } DispCnt = 3; Display_LTOF(); break;
        case 2: // display LTOF data stored in an array
          if (DispChng) {
            if (LTOF_On < 1) {VL53L0X_ON();}
            LED_Task = -30; ServoTaskPnt = -2;  // initiate scanning
            ServoAuto = false;
          } DispCnt = 6; Display_LTOF_Data(); break;
        case 3: // displays the data stored in the RangeData array
          if (DispChng) {
            if (LTOF_On < 1) {VL53L0X_ON();}
            LED_Task = -30; ServoTaskPnt = -2;  // initiate scanning
            ServoAuto = true;
          } DispCnt = 3; Display_LTOF_TgtData(); break;
        case 4: // displays the target position and range with a direction pointer
          if (DispChng) {
            if (LTOF_On < 1) {VL53L0X_ON();}
            LED_Task = -30; ServoTaskPnt = -2;  // initiate scanning
            ServoAuto = true;
          } DispCnt = 3; Display_LTOF_TgtPoint(); break;
        default: DispMode = 0; break; // over-run, returns display to DispMode = 0
      } break;
    case 3: // WiFi screens
      switch(DispMode) {
        case 0: DispCnt = 5; Display_WiFiStatus(); break;
        case 1: DispCnt = 5; Display_Wii_Ctrl(); break;
        case 2: DispCnt = 5; Display_Wii_Nunchuk(); break;
        case 3: DispCnt = 5; Display_Wii_LX16A(); break;
        default: DispMode = 0; break; // over-run, returns display to DispMode = 0
      } break;
    case 4: // IR sensor screens
      if (DispChng) {
        if (DispMode == 1) {initialiseIR();}
        if (DispMode > 2) {clearADCdata();}
      }
      switch(DispMode) {
        case 0: DispCnt = 3; Display_IRSensors(); break;
        case 1: DispCnt = 9; Display_IRLimits(); break;
        case 2: DispCnt = 3; Display_IRSensors100P(); break;
        case 3: DispCnt = 3; ADC_Rec = 0; Display_IRdata(); break;
        case 4: DispCnt = 3; ADC_Rec = 1; Display_IRdata(); break;
        case 5: DispCnt = 3; ADC_Rec = 2; Display_IRdata(); break;
        case 6: DispCnt = 3; ADC_Rec = 3; Display_IRdata(); break;
        case 7: DispCnt = 3; ADC_Rec = 4; Display_IRdata(); break;
        default: DispMode = 0; break; // over-run, returns display to DispMode = 0
      } break;
  }
  DispChng = false;
}

// --------------------------------------------------------------------------------

void Display_40ms_Sumo() {
  // DispMenu == 2 so display Sumo screens
  switch (DispPnt) {
    case 0:
      switch(DispMode) {
        case 0: DispCnt = 10; Display_Sumo_Mode0(); break;  // display sumo mode and press SW
        case 1: DispCnt =  5; Display_AIdata(); break;      // display sumo status when running
        case 2: DispCnt =  5; Display_WiFidata(); break;    // display Wi-Fi status when running
      } break;
  }
  DispChng = false;
}

// --------------------------------------------------------------------------------

void Display_40ms_Tasks() {
  // DispMenu == 1 so display task screens
  switch (DispPnt) {
    case 0:
      switch(DispMode) {
        case 0: DispCnt = 50; Display_Task_Drive(); break;
        case 1: DispCnt =  5; Display_AIdata(); break;      // display status when running
      } break;
    case 1:
      switch(DispMode) {
        case 0: DispCnt = 50; Display_Task_Hunt(); break;
        case 1: DispCnt =  5; Display_Hunt_data(); break;   // display status when running
      } break;
    case 2:
      switch(DispMode) {
        case 0: DispCnt = 50; Display_Task_Tip(); break;
        case 1: DispCnt =  5; Display_AIdata(); break;      // display status when running
      } break;
    case 3:
      switch(DispMode) {
        case 0: DispCnt = 50; Display_Task_Target(); break;
        case 1: DispCnt =  5; Display_AIdata(); break;      // display status when running
      } break;
    case 4:
      switch(DispMode) {
        case 0: DispCnt = 50; Display_Task_Distance(); break;
        case 1: DispCnt =  5; Display_AIdata(); break;      // display status when running
      } break;
  }
  DispChng = false;
}

// --------------------------------------------------------------------------------

void Display_AIdata() {
  // display sumo AI data whilst the robot is in competition mode
  Display_clear();
  // Y64 vertical text spacing, from Y=0 to Y=63
  // 0123456789012345678901234567890123456789012345678901234567890123
  //   1111111111111111    2222222222222222        3333333333333333
  Display_setFont(16);
  Display_setTextAlignment(TEXT_ALIGN_CENTER);
  Display_drawString(64, 2, AI_Title$);
  Display_drawString(64,22, AI_Status$);
  Display_drawString(64,46, String(RangeFtdTgt));
  Display_setTextAlignment(TEXT_ALIGN_LEFT);
  Display_drawString( 0,46, String(ServoAng) + Deg);
  Display_setTextAlignment(TEXT_ALIGN_RIGHT);
  Display_drawString(127,46, String(ServoAngTgtAvg) + Deg);
  Display_display();
}

// --------------------------------------------------------------------------------

void Display_Battery() {
  // display battery information
  // Y64 vertical text spacing, from Y=0 to Y=63
  // 0123456789012345678901234567890123456789012345678901234567890123
  //       1111111111111111     2222222222222222     3333333333333333
  Display_setFont(16);
  Display_setTextAlignment(TEXT_ALIGN_CENTER);
  Display_drawString(64, 06, "Battery");
  if (BatAvg > BatCritical) {
    // battery powered mode
    Display_setTextAlignment(TEXT_ALIGN_LEFT);
    Display_drawString(0, 27, String(BatAvg));
    Display_setTextAlignment(TEXT_ALIGN_CENTER);
    Display_drawString(64, 27, String(float(BatAvg)/BatCal) + "v");
    Display_setTextAlignment(TEXT_ALIGN_RIGHT);
    AnyLong = (BatAvg - BatCritical)*100/(BatMax - BatCritical);
    if (AnyLong > 100) {AnyLong = 100;} // limit max to 100%
    if (AnyLong < 0) {AnyLong = 0;}     // limit min to 0%
    Display_drawString(127, 27, String(AnyLong) + "%");
    // display the time
    Display_setTextAlignment(TEXT_ALIGN_LEFT);
    Display_drawString(0, 48, getHmsTime(millis()));
    Display_setTextAlignment(TEXT_ALIGN_RIGHT);
    Display_drawString(127, 48, BatTime$);
  } else {
    // USB powered mode
    Display_setTextAlignment(TEXT_ALIGN_CENTER);
    Display_drawString(64, 27, "USB Pwr");
    // display the time
    Display_drawString(64, 48, getHmsTime(millis()));
  }
}

// --------------------------------------------------------------------------------

void Display_BatData() {
  // plot the BatData[] array as a rolling average
  DispCX = 4;
  for (int zB = 0; zB < 60; zB++) {
    if (BatData[zB] > 0) {Display_drawRect(DispCX,63 - BatData[zB],2,BatData[zB]);}
    DispCX += 2;
  }
  // draw axice
  DispCX = 4;
  Display_drawLine(DispCX,0,DispCX,63);
  Display_drawLine(DispCX,63,DispCX + 118,63);
}

// --------------------------------------------------------------------------------

void Display_Count(int zNum) {
  // displays a number in the centre of the screen, normally a countdown
  // locks display updates to prevent display task from overwriting count
  Display_clear();
  // Y64 vertical text spacing, from Y=0 to Y=63
  // 0123456789012345678901234567890123456789012345678901234567890123
  //                     111111111111111111111111
  Display_setFont(24);
  Display_setTextAlignment(TEXT_ALIGN_CENTER);
  Display_drawString(64, 20, String(zNum));
  Display_display();
  DispClr = true; // clear memory ready for normal screen after count
  DispNOP = true; // prevent normal display updates
}

// --------------------------------------------------------------------------------

void Display_Hunt_data() {
  // display data whilst the robot is in hunt mode
  Display_clear();
  // Y64 vertical text spacing, from Y=0 to Y=63
  // 0123456789012345678901234567890123456789012345678901234567890123
  //   1111111111111111    2222222222222222        3333333333333333
  Display_setFont(16);
  Display_setTextAlignment(TEXT_ALIGN_CENTER);
  Display_drawString(64, 2, "Hunting");
  Display_drawString(64,22, String(RangeFtdTgt));
  Display_setTextAlignment(TEXT_ALIGN_LEFT);
  Display_drawString( 0,22, String(ServoAng) + Deg);
  Display_drawString( 0,46, String(L_SpdLast));
  Display_setTextAlignment(TEXT_ALIGN_RIGHT);
  Display_drawString(127,22, String(ServoAngTgtAvg) + Deg);
  Display_drawString(127,46, String(R_SpdLast));
  Display_display();
}

// --------------------------------------------------------------------------------

void Display_Init() {
  // initialise the OLED display
  Display_init();
  Display_clear();
  Display_setContrast(255);
  Display_display();
  DispClr = false; DispDisp = false;
}

// --------------------------------------------------------------------------------

void Display_Intro() {
  // first screen to be displayed
  // default fonts are ArialMT_Plain_10, ArialMT_Plain_16 and ArialMT_Plain_24
  Display_clear();
  // Y64 vertical text spacing, from Y=0 to Y=63
  // 0123456789012345678901234567890123456789012345678901234567890123
  //     111111111111111111111111        222222222222222222222222
  Display_setFont(24);
  Display_setTextAlignment(TEXT_ALIGN_CENTER);
  Display_drawString(64, 4, "SUMO");
  Display_drawString(64, 36, "ROBOT");
  Display_display();
  DispClr = true;  // display intro, then clear memory ready for first screen
}

// --------------------------------------------------------------------------------

void Display_IRLimits() {
  // display the IR sensor threshold limit values  
  // Y64 vertical text spacing, from Y=0 to Y=63
  // 0123456789012345678901234567890123456789012345678901234567890123
  //                       1111111111111111
  Display_setFont(16);
  Display_setTextAlignment(TEXT_ALIGN_CENTER);
  Display_drawString(64, 22, "SENSORS");
  // Y64 vertical text spacing, from Y=0 to Y=63
  // 0123456789012345678901234567890123456789012345678901234567890123
  // 11111111112222222222
  Display_setFont(10);
  Display_drawString(64,  0, String(AvBlk[1]));
  Display_drawString(64, 10, String(AvWht[1]));
  // 0123456789012345678901234567890123456789012345678901234567890123
  //     33333333334444444444                    55555555556666666666
  Display_setTextAlignment(TEXT_ALIGN_LEFT);
  Display_drawString(0,  4, String(AvBlk[0]));
  Display_drawString(0, 14, String(AvWht[0]));
  Display_drawString(0, 44, String(AvBlk[4]));
  Display_drawString(0, 54, String(AvWht[4]));
  // 0123456789012345678901234567890123456789012345678901234567890123
  //     77777777778888888888                    9999999999AAAAAAAAAA
  Display_setTextAlignment(TEXT_ALIGN_RIGHT);
  Display_drawString(127,  4, String(AvBlk[2]));
  Display_drawString(127, 14, String(AvWht[2]));
  Display_drawString(127, 44, String(AvBlk[3]));
  Display_drawString(127, 54, String(AvWht[3]));
  // 0123456789012345678901234567890123456789012345678901234567890123
  //                                            BBBBBBBBBB CCCCCCCCCC
  Display_setTextAlignment(TEXT_ALIGN_CENTER);
  Display_drawString(64, 43, getMaskStr());
  if (IR_BaseF) {Display_drawString(64, 54, "WHITE");}
  else {Display_drawString(64, 54, "BLACK");}
}

// --------------------------------------------------------------------------------

void Display_IRSensors() {
  // display IR sensor raw values 
  // Y64 vertical text spacing, from Y=0 to Y=63
  // 0123456789012345678901234567890123456789012345678901234567890123
  // 2222222222222222     1111111111111111
  //     3333333333333333                  4444444444444444
  Display_setFont(16);
  Display_setTextAlignment(TEXT_ALIGN_CENTER);
  Display_drawString(64, 21, "SENSORS");
  Display_drawString(64, 0, String(AvRaw[1]));    // top centre / front centre
  Display_setTextAlignment(TEXT_ALIGN_LEFT);
  Display_drawString(0, 4, String(AvRaw[0]));     // top left / front left
  Display_drawString(0, 48, String(AvRaw[4]));    // bottom left / rear left
  Display_setTextAlignment(TEXT_ALIGN_RIGHT);
  Display_drawString(127, 4, String(AvRaw[2]));   // top right / front right
  Display_drawString(127, 48, String(AvRaw[3]));  // bottom right / rear right

  // 0123456789012345678901234567890123456789012345678901234567890123
  //                                  1111111111           2222222222
  Display_setFont(10);
  Display_setTextAlignment(TEXT_ALIGN_CENTER);
  Display_drawString(64, 43, getMaskStr());
  if (IR_BaseF) {Display_drawString(64, 54, "WHITE");}
  else {Display_drawString(64, 54, "BLACK");}
}

// --------------------------------------------------------------------------------

void Display_IRSensors100P() {
  // display IR sensor values as percentages 
  // Y64 vertical text spacing, from Y=0 to Y=63
  // 0123456789012345678901234567890123456789012345678901234567890123
  // 1111111111111111     2222222222222222
  Display_setFont(16);
  Display_setTextAlignment(TEXT_ALIGN_CENTER);
  Display_drawString(64, 21, "SENSORS");
  Display_drawString(64, 0, String(Av100[1]) + "%");    // top centre / front centre
  // 0123456789012345678901234567890123456789012345678901234567890123
  //     3333333333333333                            4444444444444444
  Display_setTextAlignment(TEXT_ALIGN_LEFT);
  Display_drawString(0, 4, String(Av100[0]) + "%");     // top left / front left
  Display_drawString(0, 48, String(Av100[4]) + "%");    // bottom left / rear left
  // 0123456789012345678901234567890123456789012345678901234567890123
  //     5555555555555555                            6666666666666666
  Display_setTextAlignment(TEXT_ALIGN_RIGHT);
  Display_drawString(127, 4, String(Av100[2]) + "%");   // top right / front right
  Display_drawString(127, 48, String(Av100[3]) + "%");  // bottom right / rear right
  // 0123456789012345678901234567890123456789012345678901234567890123
  //                                            7777777777 8888888888
  Display_setFont(10);
  Display_setTextAlignment(TEXT_ALIGN_CENTER);
  Display_drawString(64, 43, getMaskStr());
  if (IR_BaseF) {Display_drawString(64, 54, "WHITE");}
  else {Display_drawString(64, 54, "BLACK");}
}

// --------------------------------------------------------------------------------

void Display_IRdata() {
  // display a sensors readings on a timeline
  // there are 40 data points so we can use 3 pixels per point on the X axis
  int zCx = 4;              // x-axis start point
  int zVP = ADC_RecPnt + 1; // oldest data point stored
  // Y64 vertical text spacing, from Y=0 to Y=63
  // 0123456789012345678901234567890123456789012345678901234567890123
  // 1111111111111111     2222222222222222
  Display_setFont(16);
  Display_setTextAlignment(TEXT_ALIGN_LEFT);
  for (int zP = 0; zP < 40; zP++) {
    if (zVP > 39) {zVP = 0;}
    int zCy = ADC_Data[zVP]/50;
    if (zCy > 0) {Display_fillRect(zCx,63-zCy,3,zCy);}
    else if (zCy == 0) {Display_fillRect(zCx, 63,3,1);}
    zVP++; zCx += 3;
  }
  Display_setColor(INVERSE);
  switch(ADC_Rec) {
    case 0: Display_drawString(0, 0, "F-L"); break; // front left
    case 1: Display_drawString(0, 0, "F-C"); break; // front centre
    case 2: Display_drawString(0, 0, "F-R"); break; // front right
    case 3: Display_drawString(0, 0, "R-R"); break; // rear right
    case 4: Display_drawString(0, 0, "R-L"); break; // rear left
  }
  Display_setColor(WHITE);
}

// --------------------------------------------------------------------------------

void Display_LTOF() {
  // display the range value from the LTOF if it is enabled
  Display_setFont(16);
  Display_setTextAlignment(TEXT_ALIGN_CENTER);

  if (LTOF_On < 1) {
    // LTOF is not active
    // Y64 vertical text spacing, from Y=0 to Y=63
    // 0123456789012345678901234567890123456789012345678901234567890123
    //       1111111111111111          2222222222222222
    Display_drawString(64, 6, "LTOF");
    Display_drawString(64,42, "Disabled");
  } else {
    // LTOF is active
    // 0123456789012345678901234567890123456789012345678901234567890123
    // 1111111111111111
    Display_drawString(64, 0, "LTOF");
    Display_setTextAlignment(TEXT_ALIGN_RIGHT);
    
    // 0123456789012345678901234567890123456789012345678901234567890123
    //                 1111111111111111
    Display_drawString(62, 16, "Range:");
    Display_setTextAlignment(TEXT_ALIGN_LEFT);
    Display_drawString(66, 16, String(Range));
    
    // 0123456789012345678901234567890123456789012345678901234567890123
    //                                 1111111111111111
    Display_setTextAlignment(TEXT_ALIGN_RIGHT);
    Display_drawString(62, 32, "Filter:");
    Display_setTextAlignment(TEXT_ALIGN_LEFT);
    Display_drawString(66, 32, String(RangeFtd));
    
    // 0123456789012345678901234567890123456789012345678901234567890123
    //                                                 1111111111111111
    Display_setTextAlignment(TEXT_ALIGN_RIGHT);
    Display_drawString(62, 48, "FPS:");
    Display_setTextAlignment(TEXT_ALIGN_LEFT);
    Display_drawString(66, 48, String(LTOF_fps));
  }
}

// --------------------------------------------------------------------------------

void Display_LTOF_Data() {
  // displays the data stored in the RangeData array
  // the table consists of 61 range values, from RangeMin to RangeOL
  int zMax = 0;
  int zMin = RangeMax;
  int zDMax = 1 + (ServoAngMaxDef - ServoAngMinDef);
  DispCX = 62 - (zDMax/2); int zXMin = DispCX;
  for (int zD = 0; zD < zDMax; zD++) {
    int zY = map(RangeData[zD],RangeMin,RangeMax,63,0);
    if (RangeData[zD] < zMin) {zMin = RangeData[zD]; zXMin = DispCX;}
    else if (RangeData[zD] > zMax) {zMax = RangeData[zD];}
    Display_drawLine(DispCX,63,DispCX,zY);
    DispCX++;
  }
  // draw the min range pointer
  Display_setColor(INVERSE);
  Display_drawLine(zXMin-1,62,zXMin-1,63);
  Display_drawLine(zXMin,61,zXMin,63);
  Display_drawLine(zXMin+1,62,zXMin+1,63);

  // Y64 vertical text spacing, from Y=0 to Y=63
  // 0123456789012345678901234567890123456789012345678901234567890123
  // 1111111111  2222222222                   33333333333 4444444444
  Display_setColor(WHITE);
  Display_setFont(10);
  Display_setTextAlignment(TEXT_ALIGN_LEFT);
  Display_drawString(0, 0, "Max");
  Display_drawString(0, 12, String(zMax));
  Display_drawString(0, 41, "Min");
  Display_drawString(0, 53, String(zMin));
  
  Display_setTextAlignment(TEXT_ALIGN_RIGHT);
  Display_drawString(127, 0, "Tgt ");
  Display_drawString(127, 12, String(ServoAngTgtAvg) + Deg);
  Display_drawString(127, 41, "Min ");
  Display_drawString(127, 53, String(ServoAngTgt) + Deg);
}

// --------------------------------------------------------------------------------

void Display_LTOF_TgtData() {
  // displays the data stored in the RangeData array, filtered by
  // the lower and upper 10mm cut-off points
  // the table consists of 61 range values, from RangeMin to RangeOL
  int zMax = 0;           // max data value
  int zMin = RangeMax;    // min data value
  int zDA = ServoAngMin;  // data point angle
  int zDMax = 1 + (ServoAngMaxDef - ServoAngMinDef);
  int zRD = RangeMax;     // modified range data varaible
  DispCX = 62 - (zDMax/2); int zXMin = DispCX;
  for (int zD = 0; zD < zDMax; zD++) {
    zRD = RangeData[zD]; zDA = zD + ServoAngMinDef;
//    if ((zDA < ServoAngTgtLL) || (zDA > ServoAngTgtUL)) {zRD = RangeMax;}
    int zY = map(zRD,RangeMin,RangeMax,63,0);
    if (RangeData[zD] < zMin) {zMin = RangeData[zD]; zXMin = DispCX;}
    else if (RangeData[zD] > zMax) {zMax = RangeData[zD];}
    Display_drawLine(DispCX,63,DispCX,zY);
    DispCX++;
  }
  // draw the min range pointer
  Display_setColor(INVERSE);
  Display_drawLine(zXMin-1,62,zXMin-1,63);
  Display_drawLine(zXMin,61,zXMin,63);
  Display_drawLine(zXMin+1,62,zXMin+1,63);

  // Y64 vertical text spacing, from Y=0 to Y=63
  // 0123456789012345678901234567890123456789012345678901234567890123
  // 1111111111  2222222222                   33333333333 4444444444
  Display_setColor(WHITE);
  Display_setFont(10);
  Display_setTextAlignment(TEXT_ALIGN_LEFT);
  Display_drawString(0, 0, "Max");
  Display_drawString(0, 12, String(zMax));
  Display_drawString(0, 41, "Min");
  Display_drawString(0, 53, String(zMin));
  
  Display_setTextAlignment(TEXT_ALIGN_RIGHT);
  Display_drawString(127, 0, "Tgt ");
  Display_drawString(127, 12, String(ServoAngTgtAvg) + Deg);
  Display_drawString(127, 41, "Min ");
  Display_drawString(127, 53, String(ServoAngTgt) + Deg);
}

// --------------------------------------------------------------------------------

void Display_LTOF_TgtPoint() {
  // displays the target position and range with a direction pointer
  // 90 Deg = 1.5708 radians
  // draw the angle pointer
  if (ServoAngTgtAvg > 90) {
    DispCX = 63 - int(63.0 * cos(0.0174533 * float(ServoAngTgtAvg)));
    DispCY = 63 - int(63.0 * cos(0.0174533 * float(ServoAngTgtAvg - 90)));
  } else if (ServoAngTgtAvg < 90) {
    DispCX = 63 - int(63.0 * cos(0.0174533 * float(ServoAngTgtAvg)));
    DispCY = 63 - int(63.0 * cos(0.0174533 * float(90 - ServoAngTgtAvg)));
  } else {
    DispCX = 64; DispCY = 0;
  }
//  PrintTx += "\nServoAngTgtAvg " + String(ServoAngTgtAvg) + "\n";
//  PrintTx += "DispCX " + String(DispCX) + "\tDispCY " + String(DispCY) + "\n";
  Display_drawLine(64,63,DispCX,DispCY);

  // draw the target at range
  Display_setColor(INVERSE);
  float zRange = (63.0 * RangeFtdTgt)/RangeMax;
  if (ServoAngTgtAvg > 90) {
    DispCX = 63 - int(zRange * cos(0.0174533 * float(ServoAngTgtAvg)));
    DispCY = 63 - int(zRange * cos(0.0174533 * float(ServoAngTgtAvg - 90)));
  } else if (ServoAngTgtAvg < 90) {
    DispCX = 63 - int(zRange * cos(0.0174533 * float(ServoAngTgtAvg)));
    DispCY = 63 - int(zRange * cos(0.0174533 * float(90 - ServoAngTgtAvg)));
  } else {
    DispCX = 64; DispCY = 63 - zRange;
  }
  Display_drawCircle(DispCX,DispCY,4);

  SpeedTgt = 1000; Task_Target_Steer(); // see the effect on normalised steering
//  display.setColor(WHITE);
  // Y64 vertical text spacing, from Y=0 to Y=63
  // 0123456789012345678901234567890123456789012345678901234567890123
  // 1111111111                          2222222222
  Display_setFont(10);
  Display_setTextAlignment(TEXT_ALIGN_LEFT);
  Display_drawString(2,  0, "L-Steer");
  Display_drawString(2, 36, "Angle");

  // 0123456789012345678901234567890123456789012345678901234567890123
  //            3333333333333333                    4444444444444444
  Display_setFont(16);
  Display_drawString(0, 11, String(L_SpdTgt));
  Display_drawString(0, 47, String(ServoAngTgtAvg) + Deg);

  // 0123456789012345678901234567890123456789012345678901234567890123
  // 1111111111                          2222222222
  Display_setTextAlignment(TEXT_ALIGN_RIGHT);
  Display_setFont(10);
  Display_drawString(127,  0, "R-Steer");
  Display_drawString(127, 36, "Range");

  // 0123456789012345678901234567890123456789012345678901234567890123
  //            3333333333333333                    4444444444444444
  Display_setFont(16);
  Display_drawString(125, 11, String(R_SpdTgt));
  Display_drawString(125, 47, String(RangeFtdTgt));
  Display_setColor(WHITE);
}

// --------------------------------------------------------------------------------

void Display_Mirrored() {
  // display robot type when DispMon is first connected
  // we clear current flags so that this is displayed immediately
  DispClr = false; DispDisp = false; DispCnt = 1;
  
  Display_clear();    // remove current display content
  Display_display();  // clears moirror display immediately
  if (Brightness != 255) {Display_setBrightness(255);}  // display may be in sleep mode
  // Y64 vertical text spacing, from Y=0 to Y=63
  // 0123456789012345678901234567890123456789012345678901234567890123
  //  111111111111111111111111 222222222222222222222222   3333333333
  Display_setFont(24);
  Display_setTextAlignment(TEXT_ALIGN_CENTER);
  Display_drawString(64, 1, "SUMO");
  Display_drawString(64, 26, "ROBOT");
  Display_setFont(10);
  Display_drawString(64,53, "Display Mirror");
  Display_display();  // display immediately
  DispClr = true;     // remove display buffer content for next display
  DispDel = 50;       // wait 2s before displaying next count
}

// --------------------------------------------------------------------------------

void Display_MPU_Acc() {
  // display MPU6050 accelerometer information
  // Y64 vertical text spacing, from Y=0 to Y=63
  // 0123456789012345678901234567890123456789012345678901234567890123
  // 1111111111111111
  Display_setFont(16);
  Display_setTextAlignment(TEXT_ALIGN_CENTER);
  Display_drawString(64, 0, "MPU Acc.");

  // 0123456789012345678901234567890123456789012345678901234567890123
  //                 1111111111111111
  Display_setTextAlignment(TEXT_ALIGN_LEFT);
  Display_drawString(0, 16, "AcX");
  Display_setTextAlignment(TEXT_ALIGN_LEFT);
  Display_drawString(36, 16, String(AcX));
  Display_setTextAlignment(TEXT_ALIGN_RIGHT);
  Display_drawString(127, 16, String(AcXAvg));

  // 0123456789012345678901234567890123456789012345678901234567890123
  //                                 1111111111111111
  Display_setTextAlignment(TEXT_ALIGN_LEFT);
  Display_drawString(0, 32, "AcY");
  Display_setTextAlignment(TEXT_ALIGN_LEFT);
  Display_drawString(36, 32, String(AcY));
  Display_setTextAlignment(TEXT_ALIGN_RIGHT);
  Display_drawString(127, 32, String(AcYAvg));

  // 0123456789012345678901234567890123456789012345678901234567890123
  //                                                 1111111111111111
  Display_setTextAlignment(TEXT_ALIGN_LEFT);
  Display_drawString(0, 48, "AcZ");
  Display_setTextAlignment(TEXT_ALIGN_LEFT);
  Display_drawString(36, 48, String(AcZ));
  Display_setTextAlignment(TEXT_ALIGN_RIGHT);
  Display_drawString(127, 48, String(AcZAvg));
}

// --------------------------------------------------------------------------------

void Display_MPU_Ang() {
  // display MPU6050 angle information
  // Y64 vertical text spacing, from Y=0 to Y=63
  // 0123456789012345678901234567890123456789012345678901234567890123
  // 1111111111111111
  Display_setFont(16);
  Display_setTextAlignment(TEXT_ALIGN_CENTER);
  Display_drawString(64, 0, "MPU Angles");

  // 0123456789012345678901234567890123456789012345678901234567890123
  //                 1111111111111111
  Display_setTextAlignment(TEXT_ALIGN_LEFT);
  Display_drawString(0, 16, "Pitch");
  Display_drawString(40, 16, String(PitchGyr,1) + Deg);
  Display_setTextAlignment(TEXT_ALIGN_RIGHT);
  Display_drawString(127, 16, String(PitchAcc,1) + Deg);

  // 0123456789012345678901234567890123456789012345678901234567890123
  //                                 1111111111111111
  Display_setTextAlignment(TEXT_ALIGN_LEFT);
  Display_drawString(0, 32, "Roll");
  Display_drawString(40, 32, String(YawGyr,1) + Deg);
  Display_setTextAlignment(TEXT_ALIGN_RIGHT);
  Display_drawString(127, 32, String(YawAcc,1) + Deg);

  // 0123456789012345678901234567890123456789012345678901234567890123
  //                                                 1111111111111111
  Display_setTextAlignment(TEXT_ALIGN_LEFT);
  Display_drawString(0, 48, "Yaw");
  Display_drawString(40, 48, String(RotGyr,1) + Deg);
}

// --------------------------------------------------------------------------------

void Display_MPU_Gyr() {
  // display MPU6050 gyroscope information
  // Y64 vertical text spacing, from Y=0 to Y=63
  // 0123456789012345678901234567890123456789012345678901234567890123
  // 1111111111111111
  Display_setFont(16);
  Display_setTextAlignment(TEXT_ALIGN_CENTER);
  Display_drawString(64, 0, "MPU Gyros");

  // 0123456789012345678901234567890123456789012345678901234567890123
  //                 1111111111111111
  Display_setTextAlignment(TEXT_ALIGN_LEFT);
  Display_drawString(0, 16, "GyX");
  Display_drawString(40, 16, String(GyX));

  // 0123456789012345678901234567890123456789012345678901234567890123
  //                                 1111111111111111
  Display_drawString(0, 32, "GyY");
  Display_drawString(40, 32, String(GyY));

  // 0123456789012345678901234567890123456789012345678901234567890123
  //                                                 1111111111111111
  Display_drawString(0, 48, "GyZ");
  Display_drawString(40, 48, String(GyZ));

  if (GyCal < 1) {
    Display_setTextAlignment(TEXT_ALIGN_RIGHT);
    Display_drawString(127, 16, String(GyXAvg));
    Display_drawString(127, 32, String(GyYAvg));
    Display_drawString(127, 48, String(GyZAvg));
  }
}

// --------------------------------------------------------------------------------

void Display_MPU_Levels() {
  // display MPU6050 angles like s spirit level
  Display_setFont(10);

  if ((fabs(PitchGyr) <= 2.0) && (fabs(YawGyr) <= 2.0)) {
    // angles are <= 2.0°
    // Y64 vertical text spacing, from Y=0 to Y=63
    // 0123456789012345678901234567890123456789012345678901234567890123
    // 1111111111                 2222222222
    Display_setTextAlignment(TEXT_ALIGN_LEFT);
    Display_drawString(72, 0, String(PitchGyr,1) + Deg);
    Display_setTextAlignment(TEXT_ALIGN_RIGHT);
    Display_drawString(24,27, String(YawGyr,1) + Deg);
    Display_drawLine(32,32,96,32);    // horizontal centre line
    DispCX = 62 + (16 * YawGyr);      // Yaw bubble
    Display_drawRect(DispCX,30,5,5);  // horizontal bubble
    Display_drawLine(64,0,64,63);     // vertical centre line
    DispCY = 30 - (16 * PitchGyr);    // Pitch bubble
    Display_drawRect(62,DispCY,5,5);  // vertical bubble
  } else {
    // angles are <= 10.0°
    if (YawGyr >= 0.0) {
      DispCX = 96; DispCY = 2 * fabs(YawGyr);
      Display_drawLine(DispCX,32 - DispCY,DispCX,32 + DispCY);  // horizontal wedge right
      Display_drawLine(32,32,DispCX,32 - DispCY);
      Display_drawLine(32,32,DispCX,32 + DispCY);
      Display_setTextAlignment(TEXT_ALIGN_LEFT);
      Display_drawString(DispCX + 4,27, String(YawGyr,1) + Deg);
    } else {
      DispCX = 32; DispCY = 2 * fabs(YawGyr);
      Display_drawLine(DispCX,32 - DispCY,DispCX,32 + DispCY);  // horizontal wedge left
      Display_drawLine(96,32,DispCX,32 - DispCY);
      Display_drawLine(96,32,DispCX,32 + DispCY);
      Display_setTextAlignment(TEXT_ALIGN_RIGHT);
      Display_drawString(DispCX - 4,27, String(YawGyr,1) + Deg);
    }
    if (PitchGyr >= 0.0) {
      DispCX = 2 * fabs(PitchGyr); DispCY = 0; 
      Display_drawLine(64 - DispCX,DispCY,64 + DispCX,DispCY);  // vertical wedge up
      Display_drawLine(64,64,64 - DispCX,DispCY);
      Display_drawLine(64,64,64 + DispCX,DispCY);
      if (YawGyr < 0.0) {
        Display_setTextAlignment(TEXT_ALIGN_LEFT);
        Display_drawString(70,53, String(PitchGyr,1) + Deg);
      } else {
        Display_setTextAlignment(TEXT_ALIGN_RIGHT);
        Display_drawString(58,53, String(PitchGyr,1) + Deg);
      }
    } else {
      DispCX = 2 * fabs(PitchGyr); DispCY = 63; 
      Display_drawLine(64 - DispCX,DispCY,64 + DispCX,DispCY);  // vertical wedge down
      Display_drawLine(64,0,64 - DispCX,DispCY);
      Display_drawLine(64,0,64 + DispCX,DispCY);
      if (YawGyr < 0.0) {
        Display_setTextAlignment(TEXT_ALIGN_LEFT);
        Display_drawString(70,0, String(PitchGyr,1) + Deg);
      } else {
        Display_setTextAlignment(TEXT_ALIGN_RIGHT);
        Display_drawString(58,0, String(PitchGyr,1) + Deg);
      }
    }
  }
}

// --------------------------------------------------------------------------------

void Display_MPU_Rot() {
  // display MPU6050 rotation angle
  // Y64 vertical text spacing, from Y=0 to Y=63
  // 0123456789012345678901234567890123456789012345678901234567890123
  //                         1111111111111111              2222222222
  Display_setFont(16);  // fonts are 10,16 or 24
  Display_setTextAlignment(TEXT_ALIGN_CENTER);
  Display_drawString(64, 24, String(RotGyr,1) + Deg);
  Display_setFont(10);
  Display_setTextAlignment(TEXT_ALIGN_LEFT);
  Display_drawString(0,  0, "Yaw");
  Display_drawString(0, 54, String(RotDif,3) + Deg);
  Display_setTextAlignment(TEXT_ALIGN_RIGHT);
  Display_drawString(127, 54, String(GyYOff));
  
  Display_drawCircle(64,32,31);
  Display_drawLine(21,32,30,32);
  Display_drawLine(97,32,106,32);
  if (RotGyr > 0.0) {
    Display_drawLine(21,32,26,21); Display_drawLine(26,21,30,32);
    Display_drawLine(97,32,102,41); Display_drawLine(102,41,106,32);
  } else if (RotGyr < 0.0) {
    Display_drawLine(21,32,26,41); Display_drawLine(26,41,30,32);
    Display_drawLine(97,32,102,21); Display_drawLine(102,21,106,32);
  }
}

// --------------------------------------------------------------------------------

void Display_Overlay_Sensors() {
  // called when changing menu modes
  // Y64 vertical text spacing, from Y=0 to Y=63
  // 0123456789012345678901234567890123456789012345678901234567890123
  //                         1111111111111111
  Display_clear();
  Display_drawRect(16,20,96,25);
  Display_setFont(16);  // fonts are 10, 16 or 24
  Display_setTextAlignment(TEXT_ALIGN_CENTER);
  Display_drawString(64, 24, "SENSORS");
  Display_setColor(WHITE);
  Display_display();
}

// --------------------------------------------------------------------------------

void Display_Overlay_Sumo() {
  // called when changing menu modes
  // Y64 vertical text spacing, from Y=0 to Y=63
  // 0123456789012345678901234567890123456789012345678901234567890123
  //                         1111111111111111
  Display_clear();
  Display_drawRect(24,20,80,25);
  Display_setFont(16);  // fonts are 10, 16 or 24
  Display_setTextAlignment(TEXT_ALIGN_CENTER);
  Display_drawString(64, 24, "SUMO");
  Display_setColor(WHITE);
  Display_display();
}

// --------------------------------------------------------------------------------

void Display_Overlay_Tasks() {
  // called when changing menu modes
  // Y64 vertical text spacing, from Y=0 to Y=63
  // 0123456789012345678901234567890123456789012345678901234567890123
  //                         1111111111111111
  Display_clear();
  Display_drawRect(24,20,80,25);
  Display_setFont(16);  // fonts are 10, 16 or 24
  Display_setTextAlignment(TEXT_ALIGN_CENTER);
  Display_drawString(64, 24, "TASKS");
  Display_setColor(WHITE);
  Display_display();
}

// --------------------------------------------------------------------------------

void Display_QText2x16(String zT0$,String zT1$) {
  // called to display two lines of text + <NO OK>, centre screen in font 16
  // Y64 vertical text spacing, from Y=0 to Y=63
  // 0123456789012345678901234567890123456789012345678901234567890123
  //     1111111111111111    22222222222222222   3333333333333333
  Display_clear();
  Display_setFont(16);  // fonts are 10, 16 or 24
  Display_setTextAlignment(TEXT_ALIGN_CENTER);
  Display_drawString(64, 4, zT0$);
  Display_drawString(64, 24, zT1$);
  Display_setTextAlignment(TEXT_ALIGN_LEFT);
  Display_drawString(0, 44, "< NO");
  Display_setTextAlignment(TEXT_ALIGN_RIGHT);
  Display_drawString(127, 44, "OK >");
  Display_display();
  Display_clear(); DispNOP = true;
}

// --------------------------------------------------------------------------------

void Display_Reboot() {
  // called when a soft reset is about to be initiated
  // Y64 vertical text spacing, from Y=0 to Y=63
  // 0123456789012345678901234567890123456789012345678901234567890123
  //           111111111111111111111111      222222222222222
  Display_clear();
  Display_setFont(24);  // fonts are 10, 16 or 24
  Display_setTextAlignment(TEXT_ALIGN_CENTER);
  Display_drawString(64, 10, "RESET");
  Display_setFont(16);
  Display_drawString(64, 40, "Pending...");
  Display_display(); DispNOP = true;
}

// --------------------------------------------------------------------------------

void Display_ServoVals() {
  // used in cdalibratio mode to display servo and filtered LTOF values 
  // Y64 vertical text spacing, from Y=0 to Y=63
  // 0123456789012345678901234567890123456789012345678901234567890123
  //       1111111111111111  2222222222222222  3333333333333333
  Display_setFont(16);
  Display_setTextAlignment(TEXT_ALIGN_RIGHT);
  Display_drawString(63,  6, "PWM:");
  Display_drawString(63, 24, "Ang:");
  Display_drawString(63, 42, "LTOF:");
  Display_setTextAlignment(TEXT_ALIGN_LEFT);
  Display_drawString(65,  6, String(servoVal));
  Display_drawString(65, 24, String(ServoAng));
  Display_drawString(65, 42, String(RangeFtd));
}

// --------------------------------------------------------------------------------

void Display_Sumo_Mode0() {
  // display Sumo Mode0 screen  
  // Y64 vertical text spacing, from Y=0 to Y=63
  // 0123456789012345678901234567890123456789012345678901234567890123
  //             1111111111111111        2222222222
  Display_setFont(16);
  Display_setTextAlignment(TEXT_ALIGN_CENTER);
  if (!WiFiEn) {Display_drawString(64, 12, "SUMO M0");}
  else {Display_drawString(64, 12, "SUMO WiFi");}
  Display_setFont(10);
  Display_drawString(64, 36, "<< PRESS SW0    ");
}

// --------------------------------------------------------------------------------

void Display_Task_Distance() {
  // display DISTANCE Task  
  // Y64 vertical text spacing, from Y=0 to Y=63
  // 0123456789012345678901234567890123456789012345678901234567890123
  //             1111111111111111        2222222222
  Display_setFont(16);
  Display_setTextAlignment(TEXT_ALIGN_CENTER);
  Display_drawString(64, 12, "DISTANCE TEST");
  Display_setFont(10);
  Display_drawString(64, 36, "<< PRESS SW0   ");
}

// --------------------------------------------------------------------------------

void Display_Task_Drive() {
  // display Hunt Task  
  // Y64 vertical text spacing, from Y=0 to Y=63
  // 0123456789012345678901234567890123456789012345678901234567890123
  //             1111111111111111        2222222222
  Display_setFont(16);
  Display_setTextAlignment(TEXT_ALIGN_CENTER);
  Display_drawString(64, 12, "DRIVE TASK");
  Display_setFont(10);
  Display_drawString(64, 36, "<< PRESS SW0   ");
}

// --------------------------------------------------------------------------------

void Display_Task_Hunt() {
  // display Hunt Task  
  // Y64 vertical text spacing, from Y=0 to Y=63
  // 0123456789012345678901234567890123456789012345678901234567890123
  //             1111111111111111        2222222222
  Display_setFont(16);
  Display_setTextAlignment(TEXT_ALIGN_CENTER);
  Display_drawString(64, 12, "HUNT TASK");
  Display_setFont(10);
  Display_drawString(64, 36, "<< PRESS SW0   ");
}

// --------------------------------------------------------------------------------

void Display_Task_Tip() {
  // display Tip Task  
  // Y64 vertical text spacing, from Y=0 to Y=63
  // 0123456789012345678901234567890123456789012345678901234567890123
  //             1111111111111111        2222222222
  Display_setFont(16);
  Display_setTextAlignment(TEXT_ALIGN_CENTER);
  Display_drawString(64, 12, "TIP TEST");
  Display_setFont(10);
  Display_drawString(64, 36, "<< PRESS SW0   ");
}

// --------------------------------------------------------------------------------

void Display_Task_Target() {
  // display Tip Task  
  // Y64 vertical text spacing, from Y=0 to Y=63
  // 0123456789012345678901234567890123456789012345678901234567890123
  //             1111111111111111        2222222222
  Display_setFont(16);
  Display_setTextAlignment(TEXT_ALIGN_CENTER);
  Display_drawString(64, 12, "TARGET TEST");
  Display_setFont(10);
  Display_drawString(64, 36, "<< PRESS SW0   ");
}

// --------------------------------------------------------------------------------

void Display_Text2x16(String zT0$,String zT1$) {
  // called to display a two lines of text centre screen in font 16
  // Y64 vertical text spacing, from Y=0 to Y=63
  // 0123456789012345678901234567890123456789012345678901234567890123
  //               1111111111111111    2222222222222222
  Display_clear();
  Display_setFont(16);  // fonts are 10,16 or 24
  Display_setTextAlignment(TEXT_ALIGN_CENTER);
  Display_drawString(64, 14, zT0$);
  Display_drawString(64, 34, zT1$);
  Display_display();
  Display_clear(); DispNOP = true;
}

// --------------------------------------------------------------------------------

void Display_Title24(String zT$) {
  // called to display a title centre screen
  // Y64 vertical text spacing, from Y=0 to Y=63
  // 0123456789012345678901234567890123456789012345678901234567890123
  //                     111111111111111111111111
  Display_clear();
  Display_setFont(24);  // fonts are ArialMT_Plain_10,16,24
  Display_setTextAlignment(TEXT_ALIGN_CENTER);
  Display_drawString(64, 20, zT$);
  Display_display();
  Display_clear(); DispNOP = true;
}

// --------------------------------------------------------------------------------

void Display_WiFidata() {
  // called when running WiFi control in competition mode
  Display_clear();
  // Y64 vertical text spacing, from Y=0 to Y=63
  // 0123456789012345678901234567890123456789012345678901234567890123
  //   1111111111111111    2222222222222222        3333333333333333
  Display_setFont(16);
  Display_setTextAlignment(TEXT_ALIGN_CENTER);
  Display_drawString(64, 2, AI_Title$);
  Display_drawString(64,22, AI_Status$);
  Display_drawString(64,46, String(RangeFtdTgt));
  Display_setTextAlignment(TEXT_ALIGN_LEFT);
  Display_drawString( 0,46, String(ServoAng) + Deg);
  Display_setTextAlignment(TEXT_ALIGN_RIGHT);
  Display_drawString(127,46, String(ServoAngTgtAvg) + Deg);
  Display_display();
}

// --------------------------------------------------------------------------------

void Display_WiFiStatus() {
  // displays WiFi link data  
  // Y64 vertical text spacing, from Y=0 to Y=63
  // 0123456789012345678901234567890123456789012345678901234567890123
  //               1111111111111111    2222222222222222
  Display_setFont(16);
  Display_setTextAlignment(TEXT_ALIGN_CENTER);
  Display_drawString(64, 0, "WiFi STATUS");

  if (WiFiConnected) {
    if (WiFiRxBytes > 0) {
      // some data has been received
      Display_setFont(10);
      Display_setTextAlignment(TEXT_ALIGN_RIGHT);
      Display_drawString(61, 18, "Rx Bytes:");
      Display_setTextAlignment(TEXT_ALIGN_LEFT);
      Display_drawString(64, 18, getNumCSV(WiFiRxBytes));
      Display_setTextAlignment(TEXT_ALIGN_RIGHT);
      Display_drawString(61, 30, "Rx Error:");
      Display_setTextAlignment(TEXT_ALIGN_LEFT);
      Display_drawString(64, 30, String(WiFiRxErr));
  
      Display_setTextAlignment(TEXT_ALIGN_RIGHT);
      Display_drawString(61, 41, "Tx Bytes:");
      Display_setTextAlignment(TEXT_ALIGN_LEFT);
      Display_drawString(64, 41, getNumCSV(WiFiTxBytes));
      Display_setTextAlignment(TEXT_ALIGN_RIGHT);
      Display_drawString(61, 53, "Tx Error:");
      Display_setTextAlignment(TEXT_ALIGN_LEFT);
      Display_drawString(64, 53, String(WiFiTxErr));
    } else {
      // transmission errors so display disconnected
      Display_setTextAlignment(TEXT_ALIGN_CENTER);
      Display_drawString(64, 22, "NO LINK");
      Display_setFont(10);
      Display_drawString(64, 48, String(myMAC));
    }
  } else {
    Display_drawString(64, 22, "Disconnected");
    Display_drawString(64, 44, "Trying " + String(WiFiTryNum));
  }
}

// --------------------------------------------------------------------------------

void Display_Wii_Ctrl() {
  // display 6 bytes received from Wii over ESP-NOW
  Display_clear();
  // display I2C values as decimal 0 - 255
  // Y64 vertical text spacing, from Y=0 to Y=63
  // 0123456789012345678901234567890123456789012345678901234567890123
  //               1111111111111111    2222222222222222
  Display_setTextAlignment(TEXT_ALIGN_RIGHT);
  Display_setFont(10); DispCX = 58;
  Display_drawString(DispCX,  0, String(RxWiFi[0]));
  Display_drawString(DispCX, 10, String(RxWiFi[1]));
  Display_drawString(DispCX, 20, String(RxWiFi[2]));
  Display_drawString(DispCX, 30, String(RxWiFi[3]));
  Display_drawString(DispCX, 40, String(RxWiFi[4]));
  Display_drawString(DispCX, 50, String(RxWiFi[5]));
  // display I2C values as binary 00000000 - 11111111
  Display_setTextAlignment(TEXT_ALIGN_LEFT); DispCX = 66;
  Display_drawString(DispCX,  0, GetBIN(RxWiFi[0]));
  Display_drawString(DispCX, 10, GetBIN(RxWiFi[1]));
  Display_drawString(DispCX, 20, GetBIN(RxWiFi[2]));
  Display_drawString(DispCX, 30, GetBIN(RxWiFi[3]));
  Display_drawString(DispCX, 40, GetBIN(RxWiFi[4]));
  Display_drawString(DispCX, 50, GetBIN(RxWiFi[5]));
  // display Wii controller type in top left corner
  Display_setFont(16);
  Display_drawString(2, 0, "Wii");
       if (WiiType == 'C') {Display_drawString(2, 18, "Csc");}
  else if (WiiType == 'N') {Display_drawString(2, 18, "Nnk");}
}

// --------------------------------------------------------------------------------

void Display_Wii_LX16A() {
  // when connected displays Wii Nunchuk joysticks and C/Z buttons
  Display_clear();
  // display I2C values as decimal 0 - 255
  // Y64 vertical text spacing, from Y=0 to Y=63
  // 0123456789012345678901234567890123456789012345678901234567890123
  //  1111111111111111
  Display_setFont(16);
  Display_setTextAlignment(TEXT_ALIGN_CENTER);
  Display_drawString(63, 1, "LX16A");
  // 0123456789012345678901234567890123456789012345678901234567890123
  //                       2222222222    3333333333    4444444444
  Display_setFont(10);
  Display_setTextAlignment(TEXT_ALIGN_RIGHT);
  Display_drawString(32, 22, "Lft=");
  Display_drawString(96, 22, "Rht=");
  Display_drawString(32, 36, "Spd=");
  Display_drawString(96, 36, "Tgt=");
  Display_drawString(32, 50, "Min=");
  Display_drawString(96, 50, "Max=");
  Display_setTextAlignment(TEXT_ALIGN_LEFT);
  Display_drawString(34, 22, String(L_Spd));
  Display_drawString(98, 22, String(R_Spd));
  Display_drawString(34, 36, String(Speed));
  Display_drawString(98, 36, String(SpeedTgt));
  Display_drawString(34, 50, String(SpeedMin));
  Display_drawString(98, 50, String(SpeedMax));
}

// --------------------------------------------------------------------------------

void Display_Wii_Nunchuk() {
  // when connected displays Wii Nunchuk joysticks and C/Z buttons
  Display_clear();
  // display I2C values as decimal 0 - 255
  // Y64 vertical text spacing, from Y=0 to Y=63
  // 0123456789012345678901234567890123456789012345678901234567890123
  //  1111111111111111
  Display_setFont(16);
  Display_setTextAlignment(TEXT_ALIGN_CENTER);
  Display_drawString(63, 1, "Nunchuk");
  // 0123456789012345678901234567890123456789012345678901234567890123
  //                       2222222222    3333333333    4444444444
  Display_setFont(10);
  Display_setTextAlignment(TEXT_ALIGN_RIGHT);
  Display_drawString(32, 22, "JX=");
  Display_drawString(96, 22, "JY=");
  Display_drawString(32, 36, "JV=");
  Display_drawString(96, 36, "JD=");
  Display_drawString(32, 50, "C=");
  Display_drawString(96, 50, "Z=");
  Display_setTextAlignment(TEXT_ALIGN_LEFT);
  Display_drawString(34, 22, String(JoyX));
  Display_drawString(98, 22, String(JoyY));
  Display_drawString(34, 36, String(JoyV));
  Display_drawString(98, 36, String(JoyD));
  Any$ = "0"; if (CZ & 2) {Any$ = "1";}
  Display_drawString(34, 50, Any$);
  Any$ = "0"; if (CZ & 1) {Any$ = "1";}
  Display_drawString(98, 50, Any$);
}

// --------------------------------------------------------------------------------

